﻿using System;
using System.Net;
using System.Windows;
// Ventuz related namespaces
using Ventuz.Remoting4;



namespace Remoting4_Demo_1
{
    // ------------------------------------------------------------------------------------------
    //
    // This Demo shows how to
    // 
    // 1. Connect to a running Ventuz instance
    // 2. Load a scene by its name
    // 3. Validate the scene to make it ready to use
    // 4. Activate the scene in a Port and make it visible in the Renderer
    // 5. Set values and trigger events on the scene via DataItems of the scene's Scene Data
    // 6. Handle Cluster and Scene status changes
    // 7. Deactivate and dispose the scene
    //
    // ------------------------------------------------------------------------------------------


    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        /// Represents the remote connection to the Ventuz machine(s).
        /// This is the central instance for Ventuz Remoting 4!
        /// </summary>
        Cluster cluster = null;
        /// <summary>
        /// Identifies the current scene loaded via Remoting 4.
        /// </summary>
        FlaggedIID sceneIID;
        /// <summary>
        /// Represents the status of the currently loaded scene.
        /// </summary>
        SceneStatus sceneStatus = null;

        public MainWindow()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Event Handler for Cluster Status change events.
        /// </summary>
        private void Cluster_StateChanged(object sender, EventArgs e)
        {
            // Cluster state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action(HandleClusterStatusUpdate));
        }

        /// <summary>
        /// Event Handler for Scene Status change events.
        /// </summary>
        private void Scene_StatusChanged(object sender, SceneStatusChangedEventArgs e)
        {
            // Scene state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action<SceneStatus>(HandleSceneStatusUpdate), e.Status);
        }

        /// <summary>
        /// Event Handler for the 'Connect' button.
        /// </summary>
        private void Connect_Clicked(object sender, RoutedEventArgs e)
        {
            // Connect to Ventuz Cluster (here only single machine)
            IPAddress ipAdr;
            if ( IPAddress.TryParse(this.tb_IP.Text, out ipAdr) )
            {
                this.cluster = new Cluster { Name = "R4_Demo1" };
                this.cluster.AddMachine(new IPEndPoint(ipAdr, Cluster.DEFAULT_PORT));
                // register event handler to receive Cluster and Scene Status updates
                this.cluster.ClusterStateChanged += Cluster_StateChanged;
                this.cluster.SceneStatusChanged += Scene_StatusChanged;
                // finally start communication
                this.cluster.Start();
            }
            else
            {
                // invalid IP Address format
                MessageBox.Show(this, string.Format("'{0}' is an invalid IP Address", this.tb_IP.Text), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event Handler for the 'Disconnect' button.
        /// </summary>
        private void Disconnect_Clicked(object sender, RoutedEventArgs e)
        {
            // Disconnect from Ventuz Cluster
            ShutDownCluster();
        }

        /// <summary>
        /// Event Handler for the 'Load/Validate/Set Active' button.
        /// </summary>
        private async void LoadValidActive_Clicked(object sender, RoutedEventArgs e)
        {
            if ( this.sceneStatus != null && R4Helper.IsValidated(this.sceneStatus.Status) )
            {
                // activate scene on Port 0 of Default Layout Scene
                await this.cluster.PortStatus(0, 0, true, this.sceneIID.IID, null, null);
            }
            else if ( this.sceneStatus != null && R4Helper.IsLoaded(this.sceneStatus.Status) )
            {
                // validate scene
                await this.cluster.Validate(this.sceneIID.IID.Value, null);
            }
            else
            {
                // load scene and get loading progress/status:
                this.bt_Load.IsEnabled = false;                             // currently loading - do not allow next step!
                this.sStatusValueLabel.Content = "Loading Scene...";        // set status text
                // load scene to Pipe 0
                this.sceneIID = await this.cluster.Load(0, this.tb_Scene.Text, LoadFlags.New | LoadFlags.Existing | LoadFlags.LoadProgress, null);

                if ( this.sceneIID.IID.HasValue )
                {
                    // Get scene status to receive SceneStatusChanged events
                    this.sceneStatus = await this.cluster.Status(this.sceneIID.IID.Value, null);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'Release' button.
        /// </summary>
        private async void Unload_Clicked(object sender, RoutedEventArgs e)
        {
            // remove Scene from its Port and release from memory
            if ( this.sceneStatus != null )
            {
                if ( this.sceneStatus.PortIndex.HasValue )
                {
                    // scene is in Port - remove first
                    await this.cluster.PortStatus(0, 0, null, null, null, null).ContinueWith((t) =>
                    {
                        // unload/dispose scene
                        this.cluster.Release(this.sceneIID.IID.Value, null);
                    });
                }
                else if ( this.sceneIID.IID.HasValue )
                {
                    // scene was not in Port - unload/dispose is sufficient
                    await this.cluster.Release(this.sceneIID.IID.Value, null);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'Send Values' button.
        /// </summary>
        private async void SendValues_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && this.sceneIID.IID.HasValue )
            {
                try
                {
                    // Set value on 'Value_1' DataItem (the properties bound to Value_1 DataItem) asynchronously.
                    // The IID of the according scene must be specified.
                    await this.cluster.DataItem(this.sceneIID.IID.Value, ".Value_1", this.value1_tb.Text, null, null);
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }

                try
                {
                    if ( !string.IsNullOrWhiteSpace(this.value2_tb.Text) )
                    {
                        await this.cluster.DataItem(this.sceneIID.IID.Value, ".Value_2", this.value2_tb.Text, null, null); 
                    }
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }

                try
                {
                    if ( !string.IsNullOrWhiteSpace(this.value3_tb.Text) )
                    {
                        await this.cluster.DataItem(this.sceneIID.IID.Value, ".Value_3", this.value3_tb.Text, null, null); 
                    }
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'Invoke Method' button.
        /// </summary>
        private async void InvokeMethod_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && this.sceneIID.IID.HasValue )
            {
                try
                {
                    // Invoke 'Jump' DataItem (the methods bound to Jump DataItem) asynchronously.
                    // The IID of the according scene must be specified.
                    // Method parameter 'value' is null because we do not need an 
                    // argument for the method invokation but it can be an interger value
                    // if bound methods are able to process arguments.
                    await this.cluster.DataItem(this.sceneIID.IID.Value, ".Jump", null, null, null);
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Window closed by user.
        /// </summary>
        private void Window_Closed(object sender, EventArgs e)
        {
            ShutDownCluster();
        }

        /// <summary>
        /// Adjust UI according to Cluster State.
        /// CALL ONLY IN MAIN UI THREAD!
        /// </summary>
        private void HandleClusterStatusUpdate()
        {
            this.cStatusValueLabel.Content = this.cluster.ClusterState;

            if ( this.cluster.ClusterState == ClusterState.Ok || (this.cluster.ClusterState & ClusterState.WarningMask) != 0 )
            {
                // OK or Warning:
                // allow disconnect and scene load
                this.bt_Connect.IsEnabled = false;
                this.bt_Disconnect.IsEnabled = true;
                this.bt_Load.IsEnabled = true;
                this.bt_Release.IsEnabled = false;
            }
            else if ( (this.cluster.ClusterState & ClusterState.ErrorMask) != 0 )
            {
                // Error state:
                // allow connect; disallow rest
                this.bt_Connect.IsEnabled = true;
                this.bt_Disconnect.IsEnabled = false;
                this.bt_Load.IsEnabled = false;
                this.bt_Load.Content = "Load";
                this.bt_Release.IsEnabled = false;
                this.bt_Send.IsEnabled = false;
                this.bt_Invoke.IsEnabled = false;
                this.sceneStatus = null;
            }
        }

        /// <summary>
        /// Adjust UI according to Scene Status.
        /// CALL ONLY IN MAIN UI THREAD!
        /// </summary>
        private void HandleSceneStatusUpdate(SceneStatus newStatus)
        {
            this.sceneStatus = newStatus;

            uint progress;
            switch ( R4Helper.GetSceneStatusValues(this.sceneStatus, out progress) )
            {
                case 0x0:               // No Status or Loading
                    if ( progress == 0 )
                    {
                        this.bt_Load.IsEnabled = true; 
                        this.bt_Load.Content = "Load";
                        this.bt_Release.IsEnabled = false;  // nothing to release!
                        this.sStatusValueLabel.Content = "No Status";
                    }
                    else
                    {
                        this.bt_Load.IsEnabled = false;     // currently loading - do not allow next step!
                        this.sStatusValueLabel.Content = string.Format("Loading Scene...{0:0.00} %", (float)progress / 2.55f);
                    }
                break;

                case 0x00000100:        // Loaded or Validating
                    if ( progress == 0 )
                    {
                        // Loaded
                        this.bt_Load.IsEnabled = true;
                        this.bt_Load.Content = "Validate";
                        this.bt_Release.IsEnabled = true;
                        this.sStatusValueLabel.Content = "Scene Loaded";

                        this.bt_Send.IsEnabled = true;
                        this.bt_Invoke.IsEnabled = true;
                    }
                    else
                    {
                        // Validating
                        this.bt_Load.IsEnabled = false;
                        this.bt_Load.Content = "Validate";
                        this.bt_Release.IsEnabled = true;
                        this.sStatusValueLabel.Content = "Validating Scene...";
                    }
                break;

                case 0x00000300:        // Validated
                    if ( this.sceneStatus.PortIndex == null )
                    {
                        // ...but not assigend to any Port / not activated
                        this.bt_Load.IsEnabled = true;
                        this.bt_Load.Content = "Set Active";
                        this.bt_Release.IsEnabled = true;
                        this.sStatusValueLabel.Content = "Validated";

                        this.bt_Send.IsEnabled = true;
                        this.bt_Invoke.IsEnabled = true;
                    }
                    else
                    {
                        // ...and assigned to a Port / activated
                        this.bt_Load.IsEnabled = false;
                        this.bt_Load.Content = "Set Active";
                        this.bt_Release.IsEnabled = true;
                        this.sStatusValueLabel.Content = "Activated";

                        this.bt_Send.IsEnabled = true;
                        this.bt_Invoke.IsEnabled = true;
                    }
                break;

                case 0x80010000:        // Scene Disposed
                    this.bt_Load.IsEnabled = true; 
                    this.bt_Load.Content = "Load";
                    this.bt_Release.IsEnabled = false;
                    this.sStatusValueLabel.Content = "Scene Disposed";

                    this.bt_Send.IsEnabled = false;
                    this.bt_Invoke.IsEnabled = false;
                break;

                case 0x80020000:        // Load Error
                    this.bt_Load.IsEnabled = true; 
                    this.bt_Load.Content = "Load";
                    this.bt_Release.IsEnabled = false;
                    this.sStatusValueLabel.Content = "Scene Load Error";

                    this.bt_Send.IsEnabled = false;
                    this.bt_Invoke.IsEnabled = false;
                break;

                default:
                    Console.WriteLine("SHOULD NOT COME HERE!!!");
                    break;
            }
        }

        /// <summary>
        /// Shuts down the connection to a Ventuz instance.
        /// </summary>
        private void ShutDownCluster()
        {
            if ( this.cluster != null )
            {
                this.cluster.Shutdown();
                // un-register event handler
                this.cluster.ClusterStateChanged -= Cluster_StateChanged;
                this.cluster.SceneStatusChanged -= Scene_StatusChanged;
                // mark connection as dead
                this.cluster = null;
            }
        }
    }

    #region Remoting 4 Helper class

    /// <summary>
    /// Remoting 4 helper class for some common tasks.
    /// </summary>
    public class R4Helper
    {
        /// <summary>
        /// Returns the SceneStatus.Status value and an optional loading/validation progress.
        /// </summary>
        public static uint GetSceneStatusValues(SceneStatus status, out uint progress)
        {
            progress = status.Status & 0xFF;
            return status.Status & 0xFFFFFF00;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Loaded'
        /// </summary>
        public static bool IsLoaded(uint status)
        {
            return (status & 0x00000100) == 0x00000100;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Validated'
        /// </summary>
        public static bool IsValidated(uint status)
        {
            return (status & 0x00000300) == 0x00000300;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Disposed'
        /// </summary>
        public static bool IsDisposed(uint status)
        {
            return (status & 0x80010000) == 0x80010000;
        }

        /// <summary>
        /// Returns true if 'status' value represents 'Error'
        /// </summary>
        public static bool IsError(uint status)
        {
            return (status & 0x80020000) == 0x80020000;
        }

        /// <summary>
        /// Returns a combined message text of the specified Exception
        /// and all of its inner Exceptions.
        /// </summary>
        public static string GetFullExceptionText(Exception ex)
        {
            if ( ex != null )
            {
                string msg = ex.Message;
                if ( ex.InnerException != null )
                    return msg + "\n" + GetFullExceptionText(ex.InnerException);
                else
                    return msg;
            }

            return string.Empty;
        }
    } 

    #endregion
}
