﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Xml.Linq;
// Ventuz related namespaces
using Ventuz.Remoting4;
using Ventuz.Remoting4.MachineService;


namespace Remoting4_Demo_2
{
    // ------------------------------------------------------------------------------------------
    //
    // This Demo shows how to
    // 
    // 1. Use the VMSDiscovery class to get a list of Ventuz machines in the local network
    // 2. Start a Ventuz Project on a multi-machine Cluster
    // 3. List and load any available scene in the configured Ventuz Project
    // 4. List the status of all loaded scenes
    // 5. Retrieve Scene Model to get Data Items for remote communication with the scene
    // 6. Set/read properties and trigger events via the according Data Items
    // 7. Retrieve and set Project Data items
    //
    // ------------------------------------------------------------------------------------------


    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        #region Properties / Members

        /// <summary>
        /// Represents the remote connection to the Ventuz machine(s).
        /// This is the central instance for Ventuz Remoting 4!
        /// </summary>
        Cluster cluster = null;

        /// <summary>
        /// Ventuz Machine Service Discovery
        /// </summary>
        VMSDiscovery vmsDiscover = null;

        /// <summary>
        /// IID of scene on which properties are set via Data Items
        /// </summary>
        IID remoteScene = IID.Invalid;

        /// <summary>
        /// IID of that scene which is active in Pipe 0 and thus rendered
        /// </summary>
        IID activeScene = IID.Invalid;

        /// <summary>
        /// List of active VMS Clients for communication with Ventuz Machines
        /// </summary>
        List<VMSClient> activeVMSClients = new List<VMSClient>();

        List<VMS> allVMachines = new List<VMS>();
        /// <summary>
        /// Contains all Ventuz Machines found by VMSDiscovery
        /// </summary>
        public List<VMS> AllVMachines
        {
            get { return this.allVMachines; }
            set
            {
                this.allVMachines = value;
                OnPropertyChanged();
            }
        }

        ObservableCollection<VMS> clusterVMachines = new ObservableCollection<VMS>();
        /// <summary>
        /// Ventuz Machines selected to build a Cluster
        /// </summary>
        public ObservableCollection<VMS> ClusterVMachines
        {
            get { return this.clusterVMachines; }
        }

        bool clusterStarted = false;
        /// <summary>
        /// Indicated that the Cluster is started.
        /// </summary>
        public bool ClusterStarted
        {
            get { return this.clusterStarted; }
            set
            {
                this.clusterStarted = value;

                OnPropertyChanged();
                OnPropertyChanged("ClusterStopped");
            }
        }

        /// <summary>
        /// Indicates that the Cluster is stopped.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool ClusterStopped
        {
            get { return !this.ClusterStarted; }
        }

        /// <summary>
        /// Indicates that machines can be added to a potential Cluster.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanAddMachines
        {
            get { return !this.ClusterStarted; }
        }

        /// <summary>
        /// Indicates that a machines can be removed from a potential Cluster.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanRemoveMachines
        {
            get { return this.clusterVMachines.Count > 0 && !this.ClusterStarted; }
        }

        /// <summary>
        /// Indicates that a Cluster can be started.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanStartCluster
        {
            get { return this.clusterVMachines.Count > 0 && !this.ClusterStarted; }
        }

        /// <summary>
        /// Indicates that a Cluster can be stopped.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanStopCluster
        {
            get { return this.clusterVMachines.Count > 0 && this.ClusterStarted; }
        }

        /// <summary>
        /// Used to notify WPF bindings that the sources of the bindings have changed.
        /// </summary>
        private void NotifyPropertyChanges()
        {
            OnPropertyChanged("CanAddMachines");
            OnPropertyChanged("CanRemoveMachines");
            OnPropertyChanged("CanStartCluster");
            OnPropertyChanged("CanStopCluster");
        }

        ObservableCollection<R4Helper.LogItem> logEntries = new ObservableCollection<R4Helper.LogItem>();
        /// <summary>
        /// Collection of log entires from Cluster machines and local Remote GUI.
        /// </summary>
        public ObservableCollection<R4Helper.LogItem> LogEntries
        {
            get { return this.logEntries; }
        }

        // TODO: use ObservableCollection?
        List<string> files = null;
        /// <summary>
        /// List of available files - here mainly: Scenes
        /// </summary>
        public List<string> Files
        {
            get { return this.files; }
        }

        ObservableCollection<R4Helper.IIDItem> loadedScenes = new ObservableCollection<R4Helper.IIDItem>();
        /// <summary>
        /// Collection of scenes loaded by the Cluster
        /// </summary>
        public ObservableCollection<R4Helper.IIDItem> LoadedScenes
        {
            get { return this.loadedScenes; }
        }

        // TODO: use ObservableCollection?
        List<R4Helper.SceneDataItem> sdProperties = null;
        /// <summary>
        /// List of Data Item properties on selected scene (remoteScene)
        /// </summary>
        public List<R4Helper.SceneDataItem> SceneDataProperties
        {
            get { return this.sdProperties; }
        }

        // TODO: use ObservableCollection?
        List<R4Helper.SceneDataItem> sdEvents = null;
        /// <summary>
        /// List of Data Item events/methods on selected scene (remoteScene)
        /// </summary>
        public List<R4Helper.SceneDataItem> SceneDataEvents
        {
            get { return this.sdEvents; }
        }

        ObservableCollection<R4Helper.FreePortInfo> freePorts = new ObservableCollection<R4Helper.FreePortInfo>();
        /// <summary>
        /// Dynamic Collection of available Port to load a scene to
        /// </summary>
        public ObservableCollection<R4Helper.FreePortInfo> FreePorts
        {
            get { return this.freePorts; }
        }

        bool loadSceneCtxEnabled = true;
        /// <summary>
        /// Property to disable certain commands on loaded scenes
        /// </summary>
        public bool LoadedScenesCtxMenuEnabled
        {
            get { return this.loadSceneCtxEnabled; }
            set
            {
                this.loadSceneCtxEnabled = value;
                OnPropertyChanged();
            }
        }

        /// <summary>
        /// If True editor is showing Project Data items, otherwise Scene Data items.
        /// </summary>
        bool isProjectData = false;

        #endregion


        #region Constructor / Close
        
        public MainWindow()
        {
            InitializeComponent();

            StartMachineDiscovery();
        }

        /// <summary>
        /// Window closed by user.
        /// </summary>
        private void Window_Closed(object sender, EventArgs e)
        {
            ShutDownCluster();
        } 

        #endregion


        #region Ventuz Machine Discovery

        /// <summary>
        /// Starts Ventuz Machine Discovery to get infos about available Ventuz Machines in the LAN
        /// </summary>
        private void StartMachineDiscovery()
        {
            if ( this.vmsDiscover == null )
            {
                this.vmsDiscover = new VMSDiscovery(10);
                this.vmsDiscover.PropertyChanged += VMSDiscovery_Changed;
                this.vmsDiscover.Discover();
            }
            else
            {
                this.vmsDiscover.Discover();
            }
        }

        /// <summary>
        /// Handles the PropertyChanged event on VMSDiscovery class
        /// </summary>
        private void VMSDiscovery_Changed(object sender, PropertyChangedEventArgs e)
        {
            // TODO: compare with existing?
            // create list from discovered Ventuz machines
            this.AllVMachines = new List<VMS>(vmsDiscover.Items);
        } 

        #endregion

        #region Cluster Control

        /// <summary>
        /// Handles adding of machine to potential Cluster
        /// </summary>
        private void AddMachineToCluster(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllMachines.SelectedItems.Count > 0 )
            {
                // add all selected items to list which defines the multi-machine Cluster.
                // double entries are handled in called method
                foreach ( var item in this.lvAllMachines.SelectedItems )
                {
                    AddToClusterList((VMS)item);
                }
            }
        }

        /// <summary>
        /// Double-click on machines adds it automatically to Cluster list
        /// </summary>
        private void MachineDoubleClick(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllMachines.SelectedItems.Count > 0 )
                AddToClusterList((VMS)this.lvAllMachines.SelectedItem);
        }

        #region AddToClusterList
        
        /// <summary>
        /// Adds IP of specified VMS instance to Cluster list if not already in list.
        /// </summary>
        private void AddToClusterList(VMS vms)
        {
            if ( this.clusterVMachines.Count == 0 )
                this.clusterVMachines.Add(vms);
            else
            {
                // check if specified machine already in cluster list
                bool contained = false;
                foreach ( VMS item in this.clusterVMachines )
                {
                    if ( item.IPAddress.Equals(vms.IPAddress) )
                    {
                        contained = true;
                        break;
                    }
                }

                if ( !contained )
                    this.clusterVMachines.Add(vms);
            }

            // Notify WPF binding to update UI elements
            NotifyPropertyChanges();
        } 

        #endregion

        /// <summary>
        /// Handles removing of machines from potential Cluster
        /// </summary>
        private void RemoveMachineFromCluster(object sender, RoutedEventArgs e)
        {
            if ( this.lvClusterMachines.SelectedItems.Count > 0 )
            {
                IList selected = this.lvClusterMachines.SelectedItems;
                for ( int i = selected.Count - 1; i >= 0; i-- )
                {
                    this.clusterVMachines.Remove((VMS)selected[i]);
                }

                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Handles Cluster start command on the list of selected Ventuz machines
        /// </summary>
        private void StartCluster(object sender, RoutedEventArgs e)
        {
            // if Cluster not running yet, a Project must be specified
            if ( string.IsNullOrWhiteSpace(this.tbProject.Text) )
            {
                LogMessage("Cannot start Cluster. No Project specified!", LogLevel.Error, true);
                return;
            }

            if ( !this.ClusterStarted && this.clusterVMachines.Count > 0 )
            {
                this.cluster = new Cluster { Name = "R4_Demo2" };
                this.activeVMSClients.Clear();

                // create Cluster and VMS Clients
                foreach ( VMS vms in this.clusterVMachines )
                {
                    this.activeVMSClients.Add(new VMSClient(vms));
                    this.cluster.AddMachine(new IPEndPoint(vms.IPAddress, Cluster.DEFAULT_PORT));
                }

                // get info about VMS clients
                List<VentuzInfo> infos = new List<VentuzInfo>();
                List<int> notRunnig = new List<int>();

                for ( int i = 0; i < this.activeVMSClients.Count; i++ )
                {
                    try
                    {
                        infos.Add(this.activeVMSClients[i].Info());
                    }
                    catch ( VMSClientException vEx )
                    {
                        System.Console.WriteLine(string.Format("Client not available: {0}", vEx.Message));  // just for info
                        notRunnig.Add(i);
                    }
                }

                // start Ventuz Runtimes with specified Project via VMS clients
                bool anyFails = false;
                bool restart = false;

                if ( notRunnig.Count != this.activeVMSClients.Count )
                {
                    // all machines are running a Ventuz application!
                    // stop and re-start with selected Project?
                    // aks what to do
                    MessageBoxResult result = MessageBox.Show("Some/All machines in the Cluster are already running a Ventuz Designer/Runtime! \nDo you want to restart them with the specified Project?", "Warning", MessageBoxButton.YesNo, MessageBoxImage.Warning);

                    if ( result == MessageBoxResult.Yes )
                    {
                        restart = true;

                        foreach ( VMSClient client in this.activeVMSClients )
                        {
                            try
                            {
                                client.Kill();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Could not stop machine [{0}]: {1}", client.IPEndPoint.Address.ToString(), ex.Message), LogLevel.Error, false);
                            }
                        }
                    }
                }
                else
                {
                    // none of the machines is running any Ventuz application
                    restart = true;
                }
                
                if ( restart )
                {
                    // try to start clients with specified Project
                    foreach ( VMSClient client in this.activeVMSClients )
                    {
                        try
                        {
                            client.Start(this.tbProject.Text);
                        }
                        catch ( Exception ex )
                        {
                            LogMessage(string.Format("Could not start machine [{0}]: {1}", client.IPEndPoint.Address.ToString(), ex.Message), LogLevel.Error, false);
                            anyFails = true;
                        }
                    }
                }

                if ( anyFails )
                {
                    LogMessage("Some Machines of the Cluster could not be started. Check the log entries!", LogLevel.Error, true);
                }
                else
                {
                    // register Event Handler...
                    this.cluster.ClusterStateChanged += Cluster_ClusterStateChanged;
                    this.cluster.SceneStatusChanged += Cluster_SceneStatusChanged;
                    this.cluster.DataItemChanged += Cluster_DataItemChanged;
                    this.cluster.LogLevel = LogLevel.Debug;
                    this.cluster.Log += Cluster_Log;
                    // ... and finally start the Cluster
                    this.cluster.Start();
                    this.ClusterStarted = true;
                }

                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Handles Cluster stop command
        /// </summary>
        private void StopCluster(object sender, RoutedEventArgs e)
        {
            if ( this.activeVMSClients.Count > 0 )
            {
                // FIRST: shut down Cluster
                this.ShutDownCluster();

                // SECOND: stop Runtimes and disconnect from VMS clients
                foreach ( VMSClient client in this.activeVMSClients )
                {
                    try
                    {
                        client.Kill();
                    }
                    catch ( Exception ex )
                    {
                        LogMessage(string.Format("Stopping Ventuz failed: {0}", ex.Message), LogLevel.Error, false);
                    }

                    client.Disconnect();
                }

                // THIRD: clear lists
                this.activeVMSClients.Clear();

                if ( this.files != null )
                {
                    this.files.Clear();
                    this.files = null; 
                }

                OnPropertyChanged("Files");
                
                this.loadedScenes.Clear();

                this.ClusterStarted = false;

                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Shuts down the connection to a Ventuz instance.
        /// </summary>
        private void ShutDownCluster()
        {
            if ( this.cluster != null )
            {
                ClearDataItemsLists();

                this.cluster.Shutdown();
                // un-register event handler
                this.cluster.ClusterStateChanged -= Cluster_ClusterStateChanged;
                this.cluster.SceneStatusChanged -= Cluster_SceneStatusChanged;
                this.cluster.DataItemChanged -= Cluster_DataItemChanged;
                this.cluster.Log -= Cluster_Log;
                // mark connection as dead
                this.cluster = null;
            }
        }

        /// <summary>
        /// Process Cluster state changes here
        /// </summary>
        private void HandleClusterStatusUpdate()
        {
            if ( this.cluster.ClusterState == ClusterState.Ok )
            {
                // OK state:
                LogMessage("Cluster is ready!", LogLevel.Info, false);
                UpdateFiles("Scenes");
                UpdateLoadedScenes();
            }
            else if ( (this.cluster.ClusterState & ClusterState.ErrorMask) != 0 )
            {
                // Error state:
                LogMessage(string.Format("Cluster is in an Error state: {0}", this.cluster.ClusterState), LogLevel.Error, false);
            }
        }

        #endregion

        #region Cluster Event Handler
        
        /// <summary>
        /// Log entry is received from any of the Cluster machines
        /// </summary>
        void Cluster_Log(object sender, LogEventArgs e)
        {
            // Cluster log events must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(() => { this.logEntries.Add(R4Helper.GetLogItem(e)); });
        }

        /// <summary>
        /// Status of any of the loaded scenes has changed
        /// </summary>
        void Cluster_SceneStatusChanged(object sender, SceneStatusChangedEventArgs e)
        {
            // Scene state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action<IID, SceneStatus>(HandleSceneStatusUpdate), e.IID, e.Status);
        }

        /// <summary>
        /// Status of the Cluster has changed
        /// </summary>
        void Cluster_ClusterStateChanged(object sender, EventArgs e)
        {
            // Cluster state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action(HandleClusterStatusUpdate));
        }

        /// <summary>
        /// Data Item changed: property value changed or event was fired
        /// </summary>
        void Cluster_DataItemChanged(object sender, DataItemChangedEventArgs e)
        {
            // Data Item change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action<string, IID, object>(HandleSceneDataChange), e.DataItem, e.IID, e.Value);
        }

        #endregion

        #region Scene Control
        
        /// <summary>
        /// Handles load commands from the context menu on a scene
        /// </summary>
        private void SceneItem_Load(object sender, RoutedEventArgs e)
        {
            MenuItem mi = sender as MenuItem;

            if ( mi != null )
            {
                if ( this.lvAllScenes.SelectedItems.Count > 0 )
                {
                    // get load modus
                    LoadFlags lf = LoadFlags.New;
                    switch ( mi.Name )
                    {
                        case "New":
                            break;
                        case "NewExisting":
                            lf |= LoadFlags.Existing;
                            break;
                        case "ExistingUnused":
                            lf = LoadFlags.Existing | LoadFlags.Unused;
                            break;
                    }

                    // load all selected scenes
                    foreach ( var item in this.lvAllScenes.SelectedItems )
                    {
                        LoadScene((string)item, lf);
                    }
                }
            }
        }

        /// <summary>
        /// Handles double-click on a scene: load existing or new instance of the selected scene
        /// </summary>
        private void SceneItem_DoubleClick(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllScenes.SelectedItems.Count > 0 )
                LoadScene((string)this.lvAllScenes.SelectedItem, LoadFlags.New | LoadFlags.Existing);
        }

        /// <summary>
        /// Sets new value of the according Data Item on the according scene
        /// </summary>
        private void SceneItemValueSet_Click(object sender, RoutedEventArgs e)
        {
            // get Data Item wrapper - it's the DataContect of the clicked Button
            R4Helper.SceneDataItem smi = ((Button)sender).DataContext as R4Helper.SceneDataItem;

            if ( smi != null )
            {
                try
                {
                    SetSceneDataValue(smi.AdressPath, smi.Value, null);
                }
                catch ( Exception ex )
                {
                    LogMessage(string.Format("Setting SceneData failed: {0}", ex.Message), LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles command to set all available input properties via Data Items
        /// </summary>
        private async void SceneItemValueSetAll_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && this.sdProperties != null )
            {
                // get Cluster Clock to schedule Data Item set commands to future
                long clock = 0;
                bool failed = true;

                while ( failed )
                {
                    try
                    {
                        clock = this.cluster.Clock(null).Result;
                        failed = false;
                    }
                    catch
                    {
                        Console.WriteLine("#### Get Cluster Clock failed! #####");  // just for Debug info
                        // failed - try again
                        failed = true;
                    }
                }
                clock += 20;    // schedule 20 frame in the future

                try
                {
                    bool setParallel = false;   // change for different Set method

                    if ( setParallel )
                    {
                        List<Task> tasks = new List<Task>();

                        if ( !this.isProjectData && !this.remoteScene.IsInvalid )
                        {
                            // Scene data items
                            foreach ( R4Helper.SceneDataItem item in this.sdProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    tasks.Add(this.cluster.DataItem(this.remoteScene, item.AdressPath, item.Value, clock, null));
                            }
                        }
                        else if ( this.isProjectData )
                        {
                            // Project data items
                            foreach ( R4Helper.SceneDataItem item in this.sdProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    tasks.Add(this.cluster.ProjectDataItem(item.AdressPath, item.Value, clock, null));
                            }
                        }

                        try
                        {
                            await Task.WhenAll(tasks);
                        }
                        catch ( Exception ex )
                        {
                            Console.WriteLine(ex.Message);
                        }
                    }
                    else
                    {
                        if ( !this.isProjectData && !this.remoteScene.IsInvalid )
                        {
                            // Scene data items
                            foreach ( R4Helper.SceneDataItem item in this.sdProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    SetSceneDataValue(item.AdressPath, item.Value, clock);
                            }
                        }
                        else if ( this.isProjectData )
                        {
                            // Project data items
                            foreach ( R4Helper.SceneDataItem item in this.sdProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    await this.cluster.ProjectDataItem(item.AdressPath, item.Value, clock, null);
                            }
                        }
                    }
                }
                catch ( Exception ex )
                {
                    LogMessage(string.Format("Setting Data Items failed: {0}", ex.Message), LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles clear command on Data Item lists
        /// </summary>
        private void DataItemsClear_Click(object sender, RoutedEventArgs e)
        {
            ClearDataItemsLists();
        }

        /// <summary>
        /// Stops polling read-only Data Items and clears the lists
        /// </summary>
        private async void ClearDataItemsLists()
        {
            StopDataItemReading();

            // wait for Data Item reading thread to stop but do not block Main Thread!
            await Task.Factory.StartNew(() =>
            {
                while ( this.readingDataItems )
                {
                    System.Threading.Thread.Sleep(10);
                }
            });

            if ( this.sdProperties != null )
            {
                this.sdProperties.Clear();
                this.sdProperties = null;
                OnPropertyChanged("SceneDataProperties");
            }

            if ( this.sdEvents != null )
            {
                this.sdEvents.Clear();
                this.sdEvents = null;
                OnPropertyChanged("SceneDataEvents");
            }
        }

        /// <summary>
        /// Sets value on Data Item specified by its adress.
        /// </summary>
        /// <param name="adress">Data Item adress</param>
        /// <param name="value">value to set via Data Item</param>
        /// <param name="when">Cluster clock time to set Data Item at; can be null</param>
        private async void SetSceneDataValue(string adress, string value, long? when)
        {
            // Set value on DataItem specified by adress asynchronously.
            // The IID of the according scene must be specified.
            if ( this.cluster != null )
            {
                try
                {
                    if ( !this.isProjectData && !this.remoteScene.IsInvalid )
                        await this.cluster.DataItem(this.remoteScene, adress, value, when, null);
                    else if ( this.isProjectData )
                        await this.cluster.ProjectDataItem(adress, value, when, null);
                }
                catch ( Exception ex )
                {
                    LogMessage(string.Format("Setting Data Item failed: {0}", ex.Message), LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles the change of a Data Item  - here only Events are processed.
        /// Output properties do not notify via Remoting about changes to reduce network traffic;
        /// their values must be polled (see StartDataItemReading() method).
        /// </summary>
        /// <param name="dataItemAdress">adress of changed Data Item</param>
        /// <param name="sceneIID">IID of scene this Data Item belongs to</param>
        /// <param name="value">new Data Item value</param>
        private void HandleSceneDataChange(string dataItemAdress, IID sceneIID, object value)
        {
            // TODO: what about the '/' ??? Events come with leading '/'; is this a BUG?
            if ( this.sdEvents != null )
            {
                dataItemAdress = dataItemAdress.Trim('/');

                foreach ( R4Helper.SceneDataItem item in this.sdEvents )
                {
                    if ( item.AdressPath == dataItemAdress )
                        item.RaiseIncomingEvent();
                } 
            }
        }

        /// <summary>
        /// Loads scene asynchronously and adds it to list of loaded scenes.
        /// </summary>
        /// <param name="scene">identity of scene to load</param>
        /// <param name="loadFlags">load modus</param>
        private async void LoadScene(string scene, LoadFlags loadFlags)
        {
            try
            {
                FlaggedIID fIID = await this.cluster.Load(0, scene, loadFlags, null);

                if ( fIID.IID.HasValue )
                {
                    // wrap and add to list of loaded scenes
                    R4Helper.IIDItem iidi = new R4Helper.IIDItem(fIID.IID.Value);

                    if ( loadFlags == LoadFlags.New )
                    {
                        // new scene instance - add to list!
                        this.loadedScenes.Add(iidi);

                        // Get scene status to receive SceneStatusChanged events
                        iidi.FullStatus = await this.cluster.Status(fIID.IID.Value, null);
                    }
                    else
                    {
                        // could be reuse of already loaded scene - check if already in list
                        bool exists = false;

                        foreach ( R4Helper.IIDItem item in this.loadedScenes )
                        {
                            if ( item.IID.Equals(iidi.IID) )
                            {
                                // found
                                exists = true;
                                break;
                            }
                        }

                        if ( !exists )
                        {
                            this.loadedScenes.Add(iidi);
                            iidi.FullStatus = await this.cluster.Status(fIID.IID.Value, null);
                        }
                    }
                }
            }
            catch ( Exception ex )
            {
                LogMessage(string.Format("Loading scene '{0}' failed: {1}", scene, ex.Message), LogLevel.Error, true);
            }
        }

        /// <summary>
        /// Handles different commands on scenes available in the Cluster
        /// </summary>
        private async void SceneCommands(object sender, RoutedEventArgs e)
        {
            MenuItem mi = sender as MenuItem;
            if ( mi != null )
            {
                if ( this.lvLoadedScenes.SelectedItems.Count > 0 )
                {
                    R4Helper.IIDItem idi = null;
                    R4Helper.FreePortInfo fpi = null;
                    string cmdName = mi.Name;

                    if ( mi.Header is R4Helper.FreePortInfo )
                    {
                        // get information about Port where the scene should be set from the Header of the MenuItem
                        cmdName = "Activate";
                        fpi = (R4Helper.FreePortInfo)mi.Header;
                    }

                    switch ( cmdName )
                    {
                        case "Validate":
                            // validate scene
                            try
                            {
                                List<Task> tasks = new List<Task>();

                                foreach ( var item in this.lvLoadedScenes.SelectedItems )
                                    tasks.Add(this.cluster.Validate(((R4Helper.IIDItem)item).IID, null));

                                await Task.WhenAll(tasks);
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Validation failed: {0}", ex.Message), LogLevel.Error, true);
                            }

                            break;
                        case "Activate":
                            // activate scene in Port of Scene specified in the FreePortInfo class
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                await this.cluster.PortStatus(fpi.Host, fpi.PortIndex, true, idi.IID, null, null);
                                this.activeScene = idi.IID;
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Activation failed: {0}", ex.Message), LogLevel.Error, true);
                            }

                            break;
                        case "Deactivate":
                            // remove scene from its assigned Port 
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                await this.cluster.PortStatus(idi.FullStatus.PortIID, idi.FullStatus.PortIndex.Value, false, IID.Invalid, null, null);
                                this.activeScene = IID.Invalid;
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Deactivation failed: {0}", ex.Message), LogLevel.Error, true);
                            }

                            break;
                        case "Release":
                            // unload/dispose scene
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                // remove from port first
                                if ( idi.FullStatus.PortIndex.HasValue )
                                    await this.cluster.PortStatus(idi.FullStatus.PortIID, idi.FullStatus.PortIndex.Value, false, IID.Invalid, null, null);
                                this.activeScene = IID.Invalid;
                                // now dispose
                                await this.cluster.Release(idi.IID, null);

                                if ( idi.IID.Equals(this.remoteScene) )
                                    ClearDataItemsLists();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Release/Dispose failed: {0}", ex.Message), LogLevel.Error, true);
                            }

                            break;
                        case "Refresh":
                            // refresh scene list 
                            this.UpdateLoadedScenes();

                            break;
                        case "Remote":
                            // make data items available in GUI for Remoting
                            try
                            {
                                StopDataItemReading();
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                this.tbSceneDITitle.Text = string.Format("Scene Data Items for: {0}", idi.Identity);
                                await ProcessSceneDataModel(idi.IID);
                                StartDataItemReading();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Presenting Scene-Data items failed: {0}", ex.Message), LogLevel.Error, true);
                            }

                            break;
                    }
                }
            }
        }

        /// <summary>
        /// Handles the request of Project Data
        /// </summary>
        private async void ProjectData_Click(object sender, RoutedEventArgs e)
        {
            await ProcessProjectDataModel();
        }

        /// <summary>
        /// Reads the Scene Model for the scene identified by the IID parameter and creates
        /// lists of properties and events which can be written or read.
        /// This provides the possibility to communicate with a remote scene.
        /// </summary>
        /// <param name="iid">IID of the scene to communicate with</param>
        private async Task ProcessSceneDataModel(IID iid)
        {
            // get Scene Model XML
            if ( this.cluster != null )
            {
                string sceneModel = await this.cluster.SceneModel(iid, null);

                this.remoteScene = iid;

                XElement root = XElement.Parse(sceneModel);
                XNamespace xns = root.GetDefaultNamespace();
                // TODO: find better/saver way to get "DataElement" node (see XPathSelectElement())
                XElement dm = root.Element(xns + "SceneModel.DataModel").Element(xns + "DataModel");

                this.isProjectData = false;

                ProcessDataModel(dm); 
            }
        }

        /// <summary>
        /// Reads the Project Model and creates
        /// lists of properties and events which can be written or read.
        /// </summary>
        private async Task ProcessProjectDataModel()
        {
            // get Project Model XML
            if ( this.cluster != null )
            {
                string projectModel = await this.cluster.DataModel(null);

                XElement root = XElement.Parse(projectModel);

                this.isProjectData = true;

                ProcessDataModel(root); 
            }
        }

        /// <summary>
        /// Creates lists of properties and events from specified XML which can be written or read.
        /// </summary>
        /// <param name="dataModel">XML which specifies the DataModel</param>
        private void ProcessDataModel(XElement dataModel)
        {
            try
            {
                // get all Data Item entries for remoting
                var leaves = from e in dataModel.Descendants() where !e.Elements().Any() select e;

                this.sdProperties = new List<R4Helper.SceneDataItem>();
                this.sdEvents = new List<R4Helper.SceneDataItem>();

                foreach ( XElement x in leaves )
                {
                    // create Data Item wrapper
                    R4Helper.SceneDataItem smi = new R4Helper.SceneDataItem(x);

                    // do not add folders to lists - this will be done later 
                    if ( smi.Itemtype == R4Helper.SceneDataItem.ItemTypes.Unknown || smi.Itemtype == R4Helper.SceneDataItem.ItemTypes.Group )
                        continue;

                    // separate properties and events
                    if ( !smi.IsTrigger )
                        this.sdProperties.Add(smi);
                    else
                        this.sdEvents.Add(smi);
                }

                // sort and group
                this.sdProperties.Sort(R4Helper.SceneDataItem.CompareByAdress);
                this.sdEvents.Sort(R4Helper.SceneDataItem.CompareByAdress);

                // create folder entries
                int pl = this.sdProperties.Count;
                Dictionary<string, string> folders = new Dictionary<string, string>();

                // ..for properties
                for ( int i = 0; i < pl; i++ )
                {
                    string folder = this.sdProperties[i].GetFolderAdress();

                    if ( folder != null )
                    {
                        string f;
                        if ( !folders.TryGetValue(folder, out f) )
                        {
                            folders.Add(folder, folder);
                            this.sdProperties.Insert(i, new R4Helper.SceneDataItem(folder));
                            i++;
                            pl++;
                        }
                    }
                }

                // .. and for events
                folders.Clear();
                pl = this.sdEvents.Count;

                for ( int i = 0; i < pl; i++ )
                {
                    string folder = this.sdEvents[i].GetFolderAdress();

                    if ( !string.IsNullOrWhiteSpace(folder) )
                    {
                        string f;
                        if ( !folders.TryGetValue(folder, out f) )
                        {
                            folders.Add(folder, folder);
                            this.sdEvents.Insert(i, new R4Helper.SceneDataItem(folder));
                            i++;
                            pl++;
                        }
                    }
                }

                // read actual values - might not match default values
                ReadCurrentDataItems();

                // WPF binding notification
                OnPropertyChanged("SceneDataProperties");
                OnPropertyChanged("SceneDataEvents");
            }
            catch ( Exception ex )
            {
                LogMessage(string.Format("Retrieving Scene Model failed: {0}", ex.Message), LogLevel.Error, true);
            }
        }

        /// <summary>
        /// Reads all Data Item values via Remoting and saves these values in the according Data Item wrappers.
        /// </summary>
        private async void ReadCurrentDataItems()
        {
            try
            {
                if ( this.sdProperties != null )
                {
                    // create Task for every Data Item read
                    List<Task<object>> readTasks = new List<Task<object>>();

                    for ( int i = 0; i < this.sdProperties.Count; i++ )
                    {
                        if ( this.sdProperties[i].ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.R) )
                            // not all entries represent Data Items - we have also folders in the list!
                            readTasks.Add(this.cluster.DataItem(this.remoteScene, this.sdProperties[i].AdressPath, null));
                    }

                    // wait until all Tasks are finished
                    await Task.WhenAll(readTasks);

                    // save returned values
                    int r = 0;  // result index
                    for ( int i = 0; i < this.sdProperties.Count; i++ )
                    {
                        if ( this.sdProperties[i].ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.R) )
                        {
                            if (  readTasks[r].Result != null )
                                this.sdProperties[i].Value = readTasks[r].Result.ToString();
                            r++;
                        }
                    }
                }
            }
            catch ( Exception ex )
            {
                LogMessage(string.Format("Get DataItems (parallel) failed: {0}", ex.Message), LogLevel.Error, false);
            }
        }

        /// <summary>
        /// Indicates that Data Items should be read (output properties)
        /// </summary>
        bool readDataItems = false;
        /// <summary>
        /// Indicates that Data Items are currently read in a background thread
        /// </summary>
        bool readingDataItems = false;

        /// <summary>
        /// Starts a background thread to read read-only Data Items (output properties)
        /// </summary>
        private void StartDataItemReading()
        {
            this.readDataItems = false;

            if ( this.sdProperties != null && this.sdProperties.Count > 0 )
            {
                Task.Factory.StartNew(() =>
                {
                    this.readDataItems = true;
                    this.readingDataItems = true;

                    try
                    {
                        while ( this.readDataItems )
                        {
                            foreach ( R4Helper.SceneDataItem item in this.sdProperties )
                            {
                                if ( !this.readDataItems )  // reading still requested
                                    break;

                                if ( item.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                                {
                                    try
                                    {
                                        // read Data Item values synchronously...
                                        object v = this.cluster.DataItem(this.remoteScene, item.AdressPath, null).Result;
                                        // ...and save in its wrapper
                                        item.Value = v.ToString();
                                    }
                                    catch ( Exception ex )
                                    {
                                        LogMessage(string.Format("Failed to read Data Item: {0}", ex.Message), LogLevel.Error, false);
                                    }

                                    if ( !this.readDataItems )  // reading still requested
                                        break;
                                }
                            }

                            // wait 500 milliseconds before starting over
                            System.Threading.Thread.Sleep(500);
                        }
                    }
                    finally
                    {
                        // signal that thread is finished
                        this.readingDataItems = false;
                    }
                }); 
            }
        }

        /// <summary>
        /// Stops the background thread for reading Data Item values
        /// </summary>
        private void StopDataItemReading()
        {
            this.readDataItems = false;
        }

        /// <summary>
        /// Searches the list of loaded scenes and updates its status.
        /// </summary>
        /// <param name="iid"></param>
        /// <param name="status"></param>
        private void HandleSceneStatusUpdate(IID iid, SceneStatus status)
        {
            // find according IIDItem
            if ( this.loadedScenes != null & this.loadedScenes.Count > 0 )
            {
                foreach ( R4Helper.IIDItem idi in this.loadedScenes )
                {
                    if ( idi.IID.Equals(iid) )
                    {
                        // ...and update its status
                        idi.FullStatus = status;
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Requests list of available files on the Cluster.
        /// </summary>
        /// <param name="datapool">Datapool that specifies the type of files</param>
        private async void UpdateFiles(string datapool)
        {
            try
            {
                string[] entries = await this.cluster.Files(datapool, null);
                this.files = new List<string>();
                this.files.AddRange(entries);
                OnPropertyChanged("Files");
            }
            catch ( Exception )
            {
                files.Clear();
                OnPropertyChanged("Files");
            }
        }

        /// <summary>
        /// Refreshes the status of loaded scenes by querying the Cluster
        /// </summary>
        private async void UpdateLoadedScenes()
        {
            try
            {
                this.loadedScenes.Clear();

                bool withLayoutScenes = true;
                IID[] scenes = await this.cluster.Scenes(0, withLayoutScenes, null);

                foreach ( IID iid in scenes )
                {
                    R4Helper.IIDItem iidi = new R4Helper.IIDItem(iid);

                    // Get scene status to receive SceneStatusChanged events
                    iidi.FullStatus = await this.cluster.Status(iid, null);

                    this.loadedScenes.Add(iidi);
                }
            }
            catch ( Exception ex )
            {
                LogMessage(string.Format("Update of scene states failed: {0}", ex.Message), LogLevel.Error, false);
            }
        }

        /// <summary>
        /// Dynamically creates sub-menu for the 'Activate' command on a loaded scene.
        /// The entries will be the Ports where the scene set to.
        /// </summary>
        private async void LoadedScenes_ContextMenuOpening(object sender, ContextMenuEventArgs e)
        {
            ListBox selectedScenes = sender as ListBox;

            if ( selectedScenes != null && selectedScenes.SelectedItems.Count > 0 )
            {
                // clear the list of free ports because it will be filled in the following
                this.freePorts.Clear();

                // disable certain commands if a LayoutScene was selected
                // TODO: does not work for active custom LayoutScenes
                R4Helper.IIDItem selIID = selectedScenes.SelectedItem as R4Helper.IIDItem;

                if ( selIID != null && selIID.FullStatus.Status == 0 )
                {
                    LoadedScenesCtxMenuEnabled = false;
                    return;
                }

                LoadedScenesCtxMenuEnabled = true;

                // get not selected items because they may have Ports where the selected scene may be set by user
                List<R4Helper.IIDItem> notSelected = new List<R4Helper.IIDItem>();
                foreach ( R4Helper.IIDItem item in this.loadedScenes )
                {
                    if ( selectedScenes.SelectedItems.Contains(item) )
                        continue;

                    if ( item is R4Helper.IIDItem )
                        notSelected.Add(item);
                }

                if ( notSelected.Count > 0 )
                {
                    // update PortInfo on the scenes
                    await Task.WhenAll(notSelected.Where((ii) => (ii.FullStatus.Status & 0x80000000) == 0).Select((item) =>
                    {
                        return cluster.Ports(item.IID, null).ContinueWith((t) =>
                        {
                            item.PortInfos = t.Result;
                            return new PortInfo[0];
                        });
                    }).ToArray());

                    // do this async. because this causes UI updates in Main Thread
                    await Task.WhenAll(notSelected.Where((ii) => (ii.FullStatus.Status & 0x80000000) == 0).Select((item) =>
                    {
                        return cluster.Status(item.IID, null).ContinueWith((t) =>
                        {
                            item.FullStatus = t.Result;
                        });
                    }).ToArray());

                    // get PipeInfo to retrieve IDs of Layoutscenes
                    int pipeCount = 0;
                    VentuzInfo? vzInfo = cluster.VentuzInfo;
                    if ( vzInfo.HasValue )
                        pipeCount = vzInfo.Value.PipeCount;

                    List<PipeInfo> pipeinfos = new List<PipeInfo>();

                    if ( pipeCount > 0 )
                    {
                        await Task.WhenAll(Enumerable.Range(0, pipeCount).Select(x =>
                        {
                            return cluster.PipeInfo(x, null).ContinueWith((t) =>
                            {
                                pipeinfos.Add(t.Result);
                            });
                        }).ToArray());
                    }

                    #region Search available Ports

                    // list of free ports per pipe:
                    // every Scene Key in the Dict. referes to an array of Port Tuples which consist of Port Index and Name
                    Dictionary<IID, R4Helper.FreePortInfo[]>    freePortsPerPipe = new Dictionary<IID, R4Helper.FreePortInfo[]>();
                    List<R4Helper.FreePortInfo[]>               freePortsPipeless = new List<R4Helper.FreePortInfo[]>();
                    bool onlyDefaultLayouts = true;

                    // first we fill with all available (not locked) ports
                    foreach ( R4Helper.IIDItem iitem in notSelected )
                    {
                        if ( (iitem.FullStatus.Status & 0x80010000) != 0 )
                            // skip disposed scenes
                            continue;

                        if ( iitem.PortInfos != null && iitem.PortInfos.Length > 0 )
                        {
                            IID pipeIdx = iitem.IID;
                            bool isLayout = false;

                            // check if this is a Pipe-LayoutScene
                            for ( int i = 0; i < pipeinfos.Count; i++ )
                            {
                                if ( pipeinfos[i].LayoutIID.Value.Equals(iitem.IID) )
                                {
                                    onlyDefaultLayouts = false;
                                    pipeIdx = (IID)i;   // adapt pipe index
                                    isLayout = true;
                                    break;
                                }
                            }

                            // search free ports
                            List<R4Helper.FreePortInfo> pis = new List<R4Helper.FreePortInfo>();
                            int pIdx = 0;

                            foreach ( PortInfo pi in iitem.PortInfos )
                            {
                                if ( (pi.Flags & PortFlags.Locked) == 0 )
                                    pis.Add(new R4Helper.FreePortInfo(pIdx++, pi.Name, iitem.IID));
                                else
                                    pIdx++; // Locked: skip index
                            }

                            // update FreePipeInfo array for Pipe-LayoutScene or normal scene with ScenePorts 
                            if ( pis.Count > 0 )
                            {
                                if ( isLayout || iitem.FullStatus.PortIndex.HasValue )
                                {
                                    if ( !freePortsPerPipe.ContainsKey(pipeIdx) )
                                        freePortsPerPipe[pipeIdx] = pis.ToArray();
                                }
                                else
                                    freePortsPipeless.Add(pis.ToArray());
                            }
                        }
                    }

                    #endregion

                    #region Create Menu Items for Free Ports

                    if ( selectedScenes.SelectedItems.Count == 1 )    // only single selection can be set to Port!
                    {
                        // check if selection is valid for setting to Port
                        bool selectionValid = true;

                        for ( int i = 0; i < pipeinfos.Count; i++ )
                        {
                            if ( pipeinfos[i].LayoutIID.Value.Equals(((R4Helper.IIDItem)selectedScenes.SelectedItems[0]).IID) )
                            {
                                // LayoutScenes not allowed to be set to Port!
                                selectionValid = false;
                                break;
                            }
                        }

                        if ( selectionValid && (freePortsPerPipe.Count > 0 || freePortsPipeless.Count > 0 || onlyDefaultLayouts) )
                        {
                            if ( onlyDefaultLayouts )
                            {
                                // create entries for default pipes
                                for ( int pi = 0; pi < pipeinfos.Count; pi++ )
                                {
                                    R4Helper.FreePortInfo fpi = new R4Helper.FreePortInfo(0, "Port 0", (IID)pi);
                                    this.freePorts.Add(fpi);
                                }
                            }

                            if ( freePortsPipeless.Count > 0 )
                            {
                                foreach ( var freePorts in freePortsPipeless )
                                {
                                    foreach ( var item in freePorts )
                                        this.freePorts.Add(item);
                                }
                            }

                            if ( freePortsPerPipe.Count > 0 )
                            {
                                foreach ( KeyValuePair<IID, R4Helper.FreePortInfo[]> freePorts in freePortsPerPipe )
                                {
                                    foreach ( var item in freePorts.Value )
                                        this.freePorts.Add(item);
                                }
                            }

                            // let's see if there are already any scenes in the available Ports
                            await Task.WhenAll(this.freePorts.Select((item) =>
                            {
                                // NOTE: setIID parameter must be IID.Invalid here!
                                return cluster.PortStatus(item.Host, item.PortIndex, null, IID.Invalid, null, null).ContinueWith((t) =>
                                {
                                    if ( t.Result.IID != null )
                                        item.SceneInPort = t.Result.IID.ToString();
                                    else
                                        item.SceneInPort = "<Free>";

                                    return item;
                                });
                            }).ToArray());
                        }
                    }

                    #endregion

                }
            }
        }

        #endregion

        #region Logging
        
        /// <summary>
        /// Handles the Clear Log contect menu command.
        /// </summary>
        private void LogClear_Click(object sender, RoutedEventArgs e)
        {
            if ( this.logEntries != null )
                this.logEntries.Clear();
        }

        /// <summary>
        /// Writes a message to the Log Window and brings up an optional Message Box to inform the user.
        /// </summary>
        /// <param name="message">message test</param>
        /// <param name="level">log/info severity</param>
        /// <param name="showPopup">true if a message box should pop up</param>
        private void LogMessage(string message, LogLevel level, bool showPopup)
        {
            if ( string.IsNullOrWhiteSpace(message) )
                return;

            if ( !this.Dispatcher.CheckAccess() )
            {
                // adding Log Entries can only be done on main thread
                Dispatcher.Invoke(new Action<string, LogLevel, bool>(LogMessage), message, level, showPopup);
                return;
            }

            R4Helper.LogItem li = new R4Helper.LogItem(level, message, "Remote Client");
            this.logEntries.Add(li);

            if ( showPopup )
            {
                string caption = "";
                MessageBoxImage icon = MessageBoxImage.Information;

                switch ( level )
                {
                    case LogLevel.Error:
                    case LogLevel.Fatal:
                        caption = "Error";
                        icon = MessageBoxImage.Error;
                        break;

                    case LogLevel.Info:
                    case LogLevel.Debug:
                    case LogLevel.None:
                        caption = "Info";
                        icon = MessageBoxImage.Information;
                        break;

                    case LogLevel.Warning:
                        caption = "Warning";
                        icon = MessageBoxImage.Warning;
                        break;

                    default:
                        break;
                }

                MessageBox.Show(this, message, caption, MessageBoxButton.OK, icon);
            }
        } 

        #endregion


        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                handler(this, new PropertyChangedEventArgs(name));
        }

        #endregion
    }

    #region class DataItemTemplateSelector

    /// <summary>
    /// This class decides which WPF Template should be used for Data Items
    /// </summary>
    public class DataItemTemplateSelector : DataTemplateSelector
    {
        public DataTemplate PropertyWriteTemplate { get; set; }
        public DataTemplate PropertyReadTemplate { get; set; }
        public DataTemplate TriggerWriteTemplate { get; set; }
        public DataTemplate TriggerReadTemplate { get; set; }
        public DataTemplate FolderTemplate { get; set; }

        public override DataTemplate SelectTemplate(object item, DependencyObject container)
        {
            R4Helper.SceneDataItem sdi = item as R4Helper.SceneDataItem;

            if ( sdi != null )
            {
                if ( sdi.IsFolder )
                    return FolderTemplate;
                else if ( !sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.RW )
                    // write properties
                    return PropertyWriteTemplate;
                else if ( !sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                    // read properties
                    return PropertyReadTemplate;
                else if ( sdi.IsTrigger && sdi.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) ) // for RW Trigger we currently use the 'Write' Template
                    // trigger methods
                    return TriggerWriteTemplate;
                else if ( sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                    // visualise events
                    return TriggerReadTemplate;
            }

            return null;
        }
    }

    #endregion

    #region Remoting 4 Helper class

    /// <summary>
    /// Remoting 4 helper class for some common tasks.
    /// </summary>
    public class R4Helper
    {
        #region inner class LogItem
        
        /// <summary>
        /// This class represents a log message entity.
        /// </summary>
        public class LogItem
        {
            /// <summary>
            /// Severity of the log entry
            /// </summary>
            public LogLevel Level { get; private set; }
            /// <summary>
            /// Log message text
            /// </summary>
            public string   Message { get; private set; }
            /// <summary>
            /// Source of the Log entry
            /// </summary>
            public string   Module { get; private set; }

            public LogItem(LogLevel level, string message, string module)
            {
                this.Level = level;
                this.Message = message;
                this.Module = module;
            }
        } 

        #endregion

        #region inner class IIDItem

        /// <summary>
        /// This class is a wrapper for the IID class to provide a convenient access to some status properties
        /// </summary>
        public class IIDItem: INotifyPropertyChanged
        {
            public IID IID { get; private set; }
            public PortInfo[] PortInfos { get; set; }

            SceneStatus fullStatus;
            public SceneStatus FullStatus
            {
                get { return fullStatus; }
                set { fullStatus = value; UpdateStatus(); }
            }

            string identity;
            public string Identity 
            { 
                get{ return identity; }
                private set { identity = value; OnPropertyChanged(); } 
            }

            string status;
            public string Status 
            { 
                get{ return status; }
                private set { status = value; OnPropertyChanged(); } 
            }

            string port;
            public string Port 
            { 
                get{ return port; }
                private set { port = value; OnPropertyChanged(); } 
            }

            private void UpdateStatus()
            {
                if ( this.fullStatus != null )
                {
                    this.Identity = this.fullStatus.Identity;
                    this.Status = R4Helper.GetSceneStatusString(this.fullStatus);

                    if ( this.fullStatus.PortIndex.HasValue )
                        this.Port = string.Format("{0}:{1}", this.fullStatus.PortIID, this.fullStatus.PortIndex.Value);
                    else
                        this.Port = "";
                }
                else
                {
                    this.Identity = "";
                    this.Status = "";
                    this.Port = "";
                }
            }

            public IIDItem(IID iid)
            {
                this.IID = iid;
                this.Identity = ""; 
                this.Port = ""; 
                this.Status = "";
             
                UpdateStatus();
            }

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }

        #endregion

        #region inner class SceneModelItem

        /// <summary>
        /// This class is a wrapper for Data Items to provide a convenient access to its properties.
        /// This is e.g. necessary for usage in WPF templates.
        /// </summary>
        public class SceneDataItem: INotifyPropertyChanged
        {
            #region enum ItemTypes
            
            /// <summary>
            /// Enum to identify according property type for a Data Item
            /// </summary>
            public enum ItemTypes
            {
                Unknown = 0,
                Group = 1,
                String = 2,
                Float = 3,
                Integer = 4,
                Boolean = 5,
                Color = 6,
                Enum = 7,
                Asset = 8,
                Trigger = 9,
                StringArray = 10,
                FloatArray = 11,
                IntegerArray = 12,
                BooleanArray = 13,
                ByteArray = 14,
            } 

            #endregion


            #region Properties

            /// <summary>
            /// Name of the Data Item. This is the last part of the remote adress.
            /// </summary>
            public string Name { get; private set; }
            /// <summary>
            /// Full adress path of this Data Item. Needed to read or write the value of this Data Item.
            /// </summary>
            public string AdressPath { get; private set; }
            /// <summary>
            /// Optional Default value of this Data Item.
            /// </summary>
            public string Default { get; private set; }

            string val = null;
            /// <summary>
            /// Actual value of this Data Item.
            /// </summary>
            public string Value
            {
                get{ return this.val; }
                set
                {
                    this.val = value;
                    OnPropertyChanged();
                }
            }

            /// <summary>
            /// Read/Write access of this Data Item
            /// </summary>
            public Ventuz.Remoting4.SceneModel.DataMode ReadWrite { get; private set; }
            /// <summary>
            /// Type of the according property/event/method of this Data Item
            /// </summary>
            public ItemTypes Itemtype { get; private set; }
            /// <summary>
            /// Indicates if this Data Item represents a Trigger/Method in the Ventuz scene.
            /// </summary>
            public bool IsTrigger { get { return this.Itemtype == ItemTypes.Trigger; } }

            /// <summary>
            /// Indicates if this instance represents only a folder but not a real Data Item
            /// </summary>
            public bool IsFolder { get; private set; }

            private int status;
            /// <summary>
            /// Property used to visualize events
            /// </summary>
            public int Status
            {
                get { return this.status;}
                set
                {
                    this.status = value;
                    OnPropertyChanged();
                }
            }

            #endregion

            #region Constructors

            public SceneDataItem(string folderPath)
            {
                this.AdressPath = folderPath;
                this.IsFolder = true;
            }

            public SceneDataItem(string name, string dflt, Ventuz.Remoting4.SceneModel.DataMode rwMode, ItemTypes type, string adress)
            {
                this.Name = name;
                this.Default = dflt;
                this.Value = dflt;
                this.ReadWrite = rwMode;
                this.Itemtype = type;
                this.AdressPath = adress;
            }

            public SceneDataItem(string name, string dflt, string rwMode, string type, string adress)
            {
                this.Name = name;
                this.Default = dflt;
                this.Value = dflt;
                this.ReadWrite = GetRWMode(rwMode);
                this.Itemtype = GetItemType(type);
                this.AdressPath = adress;
            }

            public SceneDataItem(XElement xElem)
            {
                if ( xElem == null )
                    throw new ArgumentNullException("XElement is null!");

                if ( xElem.Attribute("Name") != null )
                    this.Name = xElem.Attribute("Name").Value;
                else
                    throw new ArgumentNullException("Missing attribute Name!");

                if ( xElem.Attribute("Default") != null )   // optional
                {
                    this.Default = xElem.Attribute("Default").Value;
                    this.Value = this.Default;
                }

                this.Itemtype = GetItemType(xElem.Name.LocalName);

                this.AdressPath = GetPath(xElem);

                if ( this.Itemtype > ItemTypes.Group )
                {
                    if ( xElem.Attribute("Mode") != null )
                        this.ReadWrite = GetRWMode(xElem.Attribute("Mode").Value);
                    else
                        throw new ArgumentNullException("Missing attribute Mode!"); 
                }
            } 

            #endregion

            /// <summary>
            /// Changes Status property temporarily to visualize incoming event
            /// </summary>
            public void RaiseIncomingEvent()
            {
                this.Status = 1;
                this.Status = 0;
            }

            /// <summary>
            /// Returns the adress path of the folder this Data Item is subordinated
            /// </summary>
            public string GetFolderAdress()
            {
                int lastPIdx = this.AdressPath.LastIndexOf('.');
                string folder = string.Empty;

                if ( lastPIdx > 0 )
                    folder = this.AdressPath.Substring(1, lastPIdx - 1);

                return folder;
            }

            #region Static Helper Methods

            /// <summary>
            /// Compares two SceneDataItems by their adresses. Used for sorting.
            /// </summary>
            public static int CompareByAdress(SceneDataItem sdi1, SceneDataItem sdi2)
            {
                if ( sdi1 == null )
                {
                    if ( sdi2 == null )
                        // If both are null, they're equal.  
                        return 0;
                    else
                        // If sdi1 is null and sdi2 is not null, sdi2 is greater.  
                        return -1;
                }
                else
                {
                    // If sdi1 is not null... 
                    if ( sdi2 == null )
                        // ...and sdi2 is null, sdi1 is greater.
                        return 1;
                    else
                    {
                        // ...and sdi2 is not null, compare the adress pathes of the two SceneDataItems. 
                        int r = string.Compare(sdi1.AdressPath, sdi2.AdressPath);

                        if ( r != 0 )
                            return r;
                        else
                            // in case of same adress compare names
                            return string.Compare(sdi1.Name, sdi2.Name);
                    }
                }
            }

            /// <summary>
            /// Returns RWModes enum value from string
            /// </summary>
            private static Ventuz.Remoting4.SceneModel.DataMode GetRWMode(string rwMode)
            {
                if ( string.IsNullOrWhiteSpace(rwMode) )
                    throw new ArgumentException("Empty string does not represent an valid Read/Write Mode!");

                Ventuz.Remoting4.SceneModel.DataMode rw;
                if ( Enum.TryParse<Ventuz.Remoting4.SceneModel.DataMode>(rwMode, out rw) )
                    return rw;

                throw new ArgumentException("Invalid Read/Write Mode!");
            }

            /// <summary>
            /// Returns ItemTypes enum value from string
            /// </summary>
            private static ItemTypes GetItemType(string itemtype)
            {
                if ( string.IsNullOrWhiteSpace(itemtype) )
                    throw new ArgumentException("Empty string does not represent an valid Item Type!");

                ItemTypes it;
                if ( Enum.TryParse<ItemTypes>(itemtype, out it) )
                    return it;

                throw new ArgumentException("Invalid Item Type!");
            }

            /// <summary>
            /// Returns adress path for specified XElement
            /// </summary>
            private static string GetPath(XElement element)
            {
                XElement tmp = element;
                string path = string.Empty;

                while ( tmp != null && tmp.Attribute("Name") != null )
                {
                    path = "." + tmp.Attribute("Name").Value + path;
                    tmp = tmp.Parent;
                }

                return path;
            } 

            #endregion

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }
        
        #endregion

        #region class FreePortInfo

        /// <summary>
        /// Class that bundles information about assignable Ports
        /// </summary>
        public class FreePortInfo
        {
            public IID Host;
            public string PortName;
            public int PortIndex;
            public string SceneInPort = "";

            internal FreePortInfo(int portIndex, string portName, IID host)
            {
                Host = host;
                PortName = portName;
                PortIndex = portIndex;
            }

            public override string ToString()
            {
                return string.Format("{0} [{1}] @ {2} ({3})", PortName, PortIndex, Host, SceneInPort);
            }
        }

        #endregion

        #region static Helper Methods

        /// <summary>
        /// Returns the SceneStatus.Status value and an optional loading/validation progress.
        /// </summary>
        public static uint GetSceneStatusValues(SceneStatus status, out uint progress)
        {
            progress = status.Status & 0xFF;
            return status.Status & 0xFFFFFF00;
        }

        /// <summary>
        /// Returns the string representation of a SceneStatus.
        /// </summary>
        public static string GetSceneStatusString(SceneStatus status)
        {
            uint p;
            switch ( GetSceneStatusValues(status, out p) )
            {
                case 0x0:               // No Status or Loading
                    if ( p == 0 )
                        return "No Status";
                    else
                        return "Loading";

                case 0x00000100:        // Loaded or Validating
                    if ( p == 0 )
                        return "Loaded";
                    else
                        return "Validating";

                case 0x00000300:        // Validated
                    if ( status.PortIndex == null )
                        return "Validated";
                    else
                        return "Active";

                case 0x80010000:        // Scene Disposed
                    return "Disposed";

                case 0x80020000:        // Load Error
                    return "Load Error";

                default:
                    Console.WriteLine("SHOULD NOT COME HERE!!!");
                    break;
            }

            return "Undefined";
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Loaded'
        /// </summary>
        public static bool IsLoaded(uint status)
        {
            return (status & 0x00000100) == 0x00000100;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Validated'
        /// </summary>
        public static bool IsValidated(uint status)
        {
            return (status & 0x00000300) == 0x00000300;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Disposed'
        /// </summary>
        public static bool IsDisposed(uint status)
        {
            return (status & 0x80010000) == 0x80010000;
        }

        /// <summary>
        /// Returns true if 'status' value represents 'Error'
        /// </summary>
        public static bool IsError(uint status)
        {
            return (status & 0x80020000) == 0x80020000;
        }

        /// <summary>
        /// Returns a combined message text of the specified Exception
        /// and all of its inner Exceptions.
        /// </summary>
        public static string GetFullExceptionText(Exception ex)
        {
            if ( ex != null )
            {
                string msg = ex.Message;
                if ( ex.InnerException != null )
                    return msg + "\n" + GetFullExceptionText(ex.InnerException);
                else
                    return msg;
            }

            return string.Empty;
        }

        /// <summary>
        /// Returns a LogItem from a LogEventArgs
        /// </summary>
        public static LogItem GetLogItem(LogEventArgs le)
        {
            if ( le == null ) return null;

            return new LogItem(le.Level, le.Message, le.Module);
        }

        /// <summary>
        /// Creates SceneModel instance from the string represetation.
        /// </summary>
        /// <param name="sceneModelString">string representing the SceneModel to create</param>
        /// <param name="sceneIdentity">Identity of the scene the SceneModel belongs to</param>
        public static Ventuz.Remoting4.SceneModel.SceneModel GetSceneModel(string sceneModelString, string sceneIdentity)
        {
            if ( string.IsNullOrWhiteSpace(sceneModelString) )
                throw new ArgumentException("SceneModel string is empty!");

            if ( string.IsNullOrWhiteSpace(sceneIdentity) )
                throw new ArgumentException("Scene Identity string is empty!");

            Ventuz.Remoting4.SceneModel.SceneModel sm = null;
            using ( MemoryStream ms = new MemoryStream(Encoding.UTF8.GetBytes(sceneModelString ?? "")) )
            {
                sm = Ventuz.Remoting4.SceneModel.SceneModel.FromStream(ms, sceneIdentity);
            }

            return sm;
        }

        #endregion
    }

    #endregion
}
