﻿using System;
using System.ComponentModel;
using System.Net;
using System.Windows;
using System.Xml.Linq;
using System.IO;
using System.Text;
// Ventuz related namespaces
using Ventuz.Remoting4;
using Ventuz.Remoting4.Templates;


namespace Remoting4_Demo_3
{
    // ------------------------------------------------------------------------------------------
    //
    // This Demo shows how to
    // 
    // 1. Connect to a running Ventuz instance
    // 2. Retrieve SceneModel for scene
    // 3. Create and fill a Template
    // 4. Load Scenes by Cue and Take according Templates
    //
    // ------------------------------------------------------------------------------------------


    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        /// Represents the remote connection to the Ventuz machine(s).
        /// This is the central instance for Ventuz Remoting 4!
        /// </summary>
        Cluster cluster = null;
        /// <summary>
        /// The currently active TakeInfo associated with a Cue
        /// </summary>
        Ventuz.Remoting4.TakeInfo activeTakeInfo = null;
        /// <summary>
        /// Number of performed Takes after a Cue
        /// </summary>
        int performedTakes = 0;

        public MainWindow()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Event Handler for Cluster Status change events.
        /// </summary>
        private void Cluster_StateChanged(object sender, EventArgs e)
        {
            // Cluster state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(new Action(HandleClusterStatusUpdate));
        }

        /// <summary>
        /// Event Handler for the 'Connect' button.
        /// </summary>
        private void Connect_Clicked(object sender, RoutedEventArgs e)
        {
            // Connect to Ventuz Cluster (here only single machine)
            IPAddress ipAdr;
            if ( IPAddress.TryParse(this.tb_IP.Text, out ipAdr) )
            {
                this.cluster = new Cluster { Name = "R4_Demo3" };
                this.cluster.AddMachine(new IPEndPoint(ipAdr, Cluster.DEFAULT_PORT));
                // register event handler to receive Cluster and Scene Status updates
                this.cluster.ClusterStateChanged += Cluster_StateChanged;
                // finally start communication
                this.cluster.Start();

                // enable buttons
                this.cue_tb.IsEnabled = true;
                this.clear_tb.IsEnabled = true;
            }
            else
            {
                // invalid IP Address format
                MessageBox.Show(this, string.Format("'{0}' is an invalid IP Address", this.tb_IP.Text), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event Handler for the 'Disconnect' button.
        /// </summary>
        private void Disconnect_Clicked(object sender, RoutedEventArgs e)
        {
            // Disable buttons
            this.cue_tb.IsEnabled = false;
            this.take_tb.IsEnabled = false;
            this.clear_tb.IsEnabled = false;

            // Disconnect from Ventuz Cluster
            ShutDownCluster();
        }

        /// <summary>
        /// Event Handler for the 'Cue Scene' button.
        /// </summary>
        private async void CueScene_Clicked(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && !string.IsNullOrWhiteSpace(this.tb_Scene.Text) )
            {
                try
                {
                    // get SceneModel string from Cluster for specified scene
                    string smString = await this.cluster.SceneModel(tb_Scene.Text, null);

                    if ( string.IsNullOrWhiteSpace(smString) )
                        throw new ArgumentException(string.Format("Could not retrieve SceneModel string for: {0}", this.tb_Scene.Text));

                    // create SceneModel instance from string representation
                    Ventuz.Remoting4.SceneModel.SceneModel sm = R4Helper.GetSceneModel(smString, this.tb_Scene.Text);

                    if ( sm == null )
                        throw new ArgumentException(string.Format("Could not create SceneModel instance from SceneModel string for: {0}", this.tb_Scene.Text));

                    // create Scene Template from SceneModel
                    SceneTemplates sceneTempl = new SceneTemplates(sm);

                    // Create TemplateData for Default Template.
                    // This is important as we want to present the Scene and not a certain Animation Template.
                    TemplateData td = new TemplateData(sceneTempl.DefaultTemplate);

                    // Cue the Scene; this will load the scene
                    // The TemplateData specifies which scene should be loaded.
                    // Create new Take-ID Guid!
                    // Note: "new Guid()" will create an empty Guid which is invalid!
                    // The Take ID is necessary for further 'Takes'.
                    this.activeTakeInfo = await this.cluster.Cue(0, 0, TransitionOptions.None, td.ToString(Ventuz.Remoting4.Templates.DataFormat.JSON), Guid.NewGuid(), null, null);

                    // perform the Take to activate the scene in the ScenePort and thus the Renderer
                    if ( this.activeTakeInfo.Count > 0 )
                        await this.cluster.Take(0, 0, this.activeTakeInfo.TakeID, null, null);

                    // allow clear now
                    this.bt_Release.IsEnabled = true;  

                    // update Scene status label
                    this.sStatusValueLabel.Content = "Cued & Taken";
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'Clear Scene' button.
        /// </summary>
        private async void ClearScene_Clicked(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && !string.IsNullOrWhiteSpace(this.tb_Scene.Text) )
            {
                try
                {
                    // Cueing with 'null' data will signal that the following Take should take out the scene and clear the according ScenePort.
                    this.activeTakeInfo = await this.cluster.Cue(0, 0, TransitionOptions.None, null, Guid.NewGuid(), null, null);

                    // This Take will finally clear the ScenePort
                    await this.cluster.Take(0, 0, this.activeTakeInfo.TakeID, null, null);

                    // update Scene status label
                    this.sStatusValueLabel.Content = "Cleared";
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'CUE' button.
        /// </summary>
        private async void Cue_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && !string.IsNullOrWhiteSpace(this.tb_Scene.Text) )
            {
                try
                {
                    // get SceneModel string from Cluster for specified scene
                    string smString = await this.cluster.SceneModel(tb_Scene.Text, null);

                    if ( string.IsNullOrWhiteSpace(smString) )
                        throw new ArgumentException(string.Format("Could not retrieve SceneModel string for: {0}", this.tb_Scene.Text));

                    // create SceneModel instance from string representation
                    Ventuz.Remoting4.SceneModel.SceneModel sm = R4Helper.GetSceneModel(smString, this.tb_Scene.Text);

                    if ( sm == null )
                        throw new ArgumentException(string.Format("Could not create SceneModel instance from SceneModel string for: {0}", this.tb_Scene.Text));

                    // create Scene Template from SceneModel
                    SceneTemplates sceneTempl = new SceneTemplates(sm);

                    if ( sceneTempl.Hosts.Count > 0 )
                    {
                        // if the specified scene and its SceneTemplate contains at least one TemplateHost, we have a Template that can be cued

                        // we only cue the first available Template
                        TemplateHost th = sceneTempl.Hosts[0];

                        // create TemplateData and fill it with values
                        TemplateData td = new TemplateData(th.DefaultTemplate);

                        // set data for Template
                        // NOTE: the DataItem path must match exactly!
                        td.Data.Name = name_tb.Text;
                        td.Data.Properties.Weight = !string.IsNullOrEmpty(weight_tb.Text) ? weight_tb.Text : "0";
                        td.Data.Properties.Origin = origin_tb.Text;

                        // An alternative way to set DataItems would be this:
                        //td["Name"].SetValue(name_tb.Text);
                        //td["Properties.Weight"].SetValue(!string.IsNullOrEmpty(weight_tb.Text) ? weight_tb.Text : "0");
                        //td["Properties.Origin"].SetValue(origin_tb.Text);

                        // reset Take counter
                        this.performedTakes = 0;

                        // Cue the Template; this will load the scene with the according Template.
                        // The TemplateData specifies which scene should be loaded.
                        // Create new Take-ID Guid which will be used for the following Take commands.
                        this.activeTakeInfo = await this.cluster.Cue(0, 0, TransitionOptions.None, td.ToString(Ventuz.Remoting4.Templates.DataFormat.BIN), Guid.NewGuid(), null, null);

                        // any Takes that can be performed?
                        if ( this.activeTakeInfo.Count > 0 )
                            this.take_tb.IsEnabled = true;
                        else
                            this.take_tb.IsEnabled = false;

                        // update Scene status label
                        this.sStatusValueLabel.Content = "Cued";
                    }
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'TAKE' button.
        /// </summary>
        private async void Take_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && this.activeTakeInfo != null )
            {
                try
                {
                    if ( this.activeTakeInfo.Count > 0 && this.performedTakes < this.activeTakeInfo.Count )
                    {
                        // perform Take for TakeID specified in TakeInfo
                        await this.cluster.Take(0, 0, this.activeTakeInfo.TakeID, null, null);
                        this.performedTakes++;

                        // update Scene status label
                        this.sStatusValueLabel.Content = string.Format("Take {0}", this.performedTakes);

                        // disable TAKE button if there are no more takes to perform
                        if ( this.performedTakes >= this.activeTakeInfo.Count )
                            this.take_tb.IsEnabled = false;
                    }
                }
                catch ( Exception ex )
                {
                    string msg = R4Helper.GetFullExceptionText(ex);
                    MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        /// <summary>
        /// Event Handler for the 'CLEAR' button.
        /// </summary>
        private async void Clear_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                if ( this.cluster != null )
                {
                    // Cueing with 'null' data will signal that the following Take should take out the scene and clear the according ScenePort.
                    this.activeTakeInfo = await this.cluster.Cue(0, 0, TransitionOptions.None, null, Guid.NewGuid(), null, null);

                    // This Take will finally clear the ScenePort
                    await this.cluster.Take(0, 0, this.activeTakeInfo.TakeID, null, null);

                    // update Scene status label
                    this.sStatusValueLabel.Content = "Cleared";
                }
            }
            catch ( Exception ex )
            {
                string msg = R4Helper.GetFullExceptionText(ex);
                MessageBox.Show(this, msg, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Window closed by user.
        /// </summary>
        private void Window_Closed(object sender, EventArgs e)
        {
            ShutDownCluster();
        }

        /// <summary>
        /// Adjust UI according to Cluster State.
        /// CALL ONLY IN MAIN UI THREAD!
        /// </summary>
        private void HandleClusterStatusUpdate()
        {
            this.cStatusValueLabel.Content = this.cluster.ClusterState;

            if ( this.cluster.ClusterState == ClusterState.Ok || (this.cluster.ClusterState & ClusterState.WarningMask) != 0 )
            {
                // OK or Warning:
                // allow disconnect and scene load
                this.bt_Connect.IsEnabled = false;
                this.bt_Disconnect.IsEnabled = true;
                this.bt_Load.IsEnabled = true;
                this.bt_Release.IsEnabled = false;
            }
            else if ( (this.cluster.ClusterState & ClusterState.ErrorMask) != 0 )
            {
                // Error state:
                // allow connect; disallow rest
                this.bt_Connect.IsEnabled = true;
                this.bt_Disconnect.IsEnabled = false;
                this.bt_Load.IsEnabled = false;
                this.bt_Release.IsEnabled = false;
            }
        }

        /// <summary>
        /// Shuts down the connection to a Ventuz instance.
        /// </summary>
        private void ShutDownCluster()
        {
            if ( this.cluster != null )
            {
                this.cluster.Shutdown();
                // un-register event handler
                this.cluster.ClusterStateChanged -= Cluster_StateChanged;
                // mark connection as dead
                this.cluster = null;
            }
        }
    }


    #region Remoting 4 Helper class

    /// <summary>
    /// Remoting 4 helper class for some common tasks.
    /// </summary>
    public class R4Helper
    {
        #region inner class LogItem
        
        /// <summary>
        /// This class represents a log message entity.
        /// </summary>
        public class LogItem
        {
            /// <summary>
            /// Severity of the log entry
            /// </summary>
            public LogLevel Level { get; private set; }
            /// <summary>
            /// Log message text
            /// </summary>
            public string   Message { get; private set; }
            /// <summary>
            /// Source of the Log entry
            /// </summary>
            public string   Module { get; private set; }

            public LogItem(LogLevel level, string message, string module)
            {
                this.Level = level;
                this.Message = message;
                this.Module = module;
            }
        } 

        #endregion

        #region inner class IIDItem

        /// <summary>
        /// This class is a wrapper for the IID class to provide a convenient access to some status properties
        /// </summary>
        public class IIDItem: INotifyPropertyChanged
        {
            public IID IID { get; private set; }
            public PortInfo[] PortInfos { get; set; }

            SceneStatus fullStatus;
            public SceneStatus FullStatus
            {
                get { return fullStatus; }
                set { fullStatus = value; UpdateStatus(); }
            }

            string identity;
            public string Identity 
            { 
                get{ return identity; }
                private set { identity = value; OnPropertyChanged(); } 
            }

            string status;
            public string Status 
            { 
                get{ return status; }
                private set { status = value; OnPropertyChanged(); } 
            }

            string port;
            public string Port 
            { 
                get{ return port; }
                private set { port = value; OnPropertyChanged(); } 
            }

            private void UpdateStatus()
            {
                if ( this.fullStatus != null )
                {
                    this.Identity = this.fullStatus.Identity;
                    this.Status = R4Helper.GetSceneStatusString(this.fullStatus);

                    if ( this.fullStatus.PortIndex.HasValue )
                        this.Port = string.Format("{0}:{1}", this.fullStatus.PortIID, this.fullStatus.PortIndex.Value);
                    else
                        this.Port = "";
                }
                else
                {
                    this.Identity = "";
                    this.Status = "";
                    this.Port = "";
                }
            }

            public IIDItem(IID iid)
            {
                this.IID = iid;
                this.Identity = ""; 
                this.Port = ""; 
                this.Status = "";
             
                UpdateStatus();
            }

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }

        #endregion

        #region inner class SceneModelItem

        /// <summary>
        /// This class is a wrapper for Data Items to provide a convenient access to its properties.
        /// This is e.g. necessary for usage in WPF templates.
        /// </summary>
        public class SceneDataItem: INotifyPropertyChanged
        {
            #region enum RWModes
            
            /// <summary>
            /// Read/Write modes for Data Items
            /// </summary>
            [Flags]
            public enum RWModes
            {
                /// <summary>
                /// Not specified
                /// </summary>
                None = 0,
                /// <summary>
                /// Read-Only access
                /// </summary>
                R = 1,
                /// <summary>
                /// Write-Only access
                /// </summary>
                W = 2,
                /// <summary>
                /// Read and Write access
                /// </summary>
                RW = R | W,
            } 

            #endregion

            #region enum ItemTypes
            
            /// <summary>
            /// Enum to identify according property type for a Data Item
            /// </summary>
            public enum ItemTypes
            {
                Unknown = 0,
                Group = 1,
                String = 2,
                Float = 3,
                Integer = 4,
                Boolean = 5,
                Color = 6,
                Enum = 7,
                Asset = 8,
                Trigger = 9,
                StringArray = 10,
                FloatArray = 11,
                IntegerArray = 12,
                BooleanArray = 13,
                ByteArray = 14,
            } 

            #endregion


            #region Properties

            /// <summary>
            /// Name of the Data Item. This is the last part of the remote adress.
            /// </summary>
            public string Name { get; private set; }
            /// <summary>
            /// Full adress path of this Data Item. Needed to read or write the value of this Data Item.
            /// </summary>
            public string AdressPath { get; private set; }
            /// <summary>
            /// Optional Default value of this Data Item.
            /// </summary>
            public string Default { get; private set; }

            string val = null;
            /// <summary>
            /// Actual value of this Data Item.
            /// </summary>
            public string Value
            {
                get{ return this.val; }
                set
                {
                    this.val = value;
                    OnPropertyChanged();
                }
            }

            /// <summary>
            /// Read/Write access of this Data Item
            /// </summary>
            public RWModes ReadWrite { get; private set; }
            /// <summary>
            /// Type of the according property/event/method of this Data Item
            /// </summary>
            public ItemTypes Itemtype { get; private set; }
            /// <summary>
            /// Indicates if this Data Item represents a Trigger/Method in the Ventuz scene.
            /// </summary>
            public bool IsTrigger { get { return this.Itemtype == ItemTypes.Trigger; } }

            /// <summary>
            /// Indicates if this instance represents only a folder but not a real Data Item
            /// </summary>
            public bool IsFolder { get; private set; }

            private int status;
            /// <summary>
            /// Property used to visualize events
            /// </summary>
            public int Status
            {
                get { return this.status;}
                set
                {
                    this.status = value;
                    OnPropertyChanged();
                }
            }

            #endregion

            #region Constructors

            public SceneDataItem(string folderPath)
            {
                this.AdressPath = folderPath;
                this.IsFolder = true;
            }

            public SceneDataItem(string name, string dflt, RWModes rwMode, ItemTypes type, string adress)
            {
                this.Name = name;
                this.Default = dflt;
                this.Value = dflt;
                this.ReadWrite = rwMode;
                this.Itemtype = type;
                this.AdressPath = adress;
            }

            public SceneDataItem(string name, string dflt, string rwMode, string type, string adress)
            {
                this.Name = name;
                this.Default = dflt;
                this.Value = dflt;
                this.ReadWrite = GetRWMode(rwMode);
                this.Itemtype = GetItemType(type);
                this.AdressPath = adress;
            }

            public SceneDataItem(XElement xElem)
            {
                if ( xElem == null )
                    throw new ArgumentNullException("XElement is null!");

                if ( xElem.Attribute("Name") != null )
                    this.Name = xElem.Attribute("Name").Value;
                else
                    throw new ArgumentNullException("Missing attribute Name!");

                if ( xElem.Attribute("Default") != null )   // optional
                {
                    this.Default = xElem.Attribute("Default").Value;
                    this.Value = this.Default;
                }

                this.Itemtype = GetItemType(xElem.Name.LocalName);

                this.AdressPath = GetPath(xElem);

                if ( this.Itemtype > ItemTypes.Group )
                {
                    if ( xElem.Attribute("Mode") != null )
                        this.ReadWrite = GetRWMode(xElem.Attribute("Mode").Value);
                    else
                        throw new ArgumentNullException("Missing attribute Mode!"); 
                }
            } 

            #endregion

            /// <summary>
            /// Changes Status property temporarily to visualize incoming event
            /// </summary>
            public void RaiseIncomingEvent()
            {
                this.Status = 1;
                this.Status = 0;
            }

            /// <summary>
            /// Returns the adress path of the folder this Data Item is subordinated
            /// </summary>
            public string GetFolderAdress()
            {
                int lastPIdx = this.AdressPath.LastIndexOf('.');
                string folder = string.Empty;

                if ( lastPIdx > 0 )
                    folder = this.AdressPath.Substring(1, lastPIdx - 1);

                return folder;
            }

            #region Static Helper Methods

            /// <summary>
            /// Compares two SceneDataItems by their adresses. Used for sorting.
            /// </summary>
            public static int CompareByAdress(SceneDataItem sdi1, SceneDataItem sdi2)
            {
                if ( sdi1 == null )
                {
                    if ( sdi2 == null )
                        // If both are null, they're equal.  
                        return 0;
                    else
                        // If sdi1 is null and sdi2 is not null, sdi2 is greater.  
                        return -1;
                }
                else
                {
                    // If sdi1 is not null... 
                    if ( sdi2 == null )
                        // ...and sdi2 is null, sdi1 is greater.
                        return 1;
                    else
                    {
                        // ...and sdi2 is not null, compare the adress pathes of the two SceneDataItems. 
                        int r = string.Compare(sdi1.AdressPath, sdi2.AdressPath);

                        if ( r != 0 )
                            return r;
                        else
                            // in case of same adress compare names
                            return string.Compare(sdi1.Name, sdi2.Name);
                    }
                }
            }

            /// <summary>
            /// Returns RWModes enum value from string
            /// </summary>
            private static RWModes GetRWMode(string rwMode)
            {
                if ( string.IsNullOrWhiteSpace(rwMode) )
                    throw new ArgumentException("Empty string does not represent an valid Read/Write Mode!");

                RWModes rw;
                if ( Enum.TryParse<RWModes>(rwMode, out rw) )
                    return rw;

                throw new ArgumentException("Invalid Read/Write Mode!");
            }

            /// <summary>
            /// Returns ItemTypes enum value from string
            /// </summary>
            private static ItemTypes GetItemType(string itemtype)
            {
                if ( string.IsNullOrWhiteSpace(itemtype) )
                    throw new ArgumentException("Empty string does not represent an valid Item Type!");

                ItemTypes it;
                if ( Enum.TryParse<ItemTypes>(itemtype, out it) )
                    return it;

                throw new ArgumentException("Invalid Item Type!");
            }

            /// <summary>
            /// Returns adress path for specified XElement
            /// </summary>
            private static string GetPath(XElement element)
            {
                XElement tmp = element;
                string path = string.Empty;

                while ( tmp != null && tmp.Attribute("Name") != null )
                {
                    path = "." + tmp.Attribute("Name").Value + path;
                    tmp = tmp.Parent;
                }

                return path;
            } 

            #endregion

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }
        
        #endregion

        #region class FreePortInfo

        /// <summary>
        /// Class that bundles information about assignable Ports
        /// </summary>
        public class FreePortInfo
        {
            public IID Host;
            public string PortName;
            public int PortIndex;

            internal FreePortInfo(int portIndex, string portName, IID host)
            {
                Host = host;
                PortName = portName;
                PortIndex = portIndex;
            }

            public override string ToString()
            {
                return string.Format("{0} [{1}] @ {2}", PortName, PortIndex, Host);
            }
        }

        #endregion

        #region static Helper Methods

        /// <summary>
        /// Returns the SceneStatus.Status value and an optional loading/validation progress.
        /// </summary>
        public static uint GetSceneStatusValues(SceneStatus status, out uint progress)
        {
            progress = status.Status & 0xFF;
            return status.Status & 0xFFFFFF00;
        }

        /// <summary>
        /// Returns the string representation of a SceneStatus.
        /// </summary>
        public static string GetSceneStatusString(SceneStatus status)
        {
            uint p;
            switch ( GetSceneStatusValues(status, out p) )
            {
                case 0x0:               // No Status or Loading
                    if ( p == 0 )
                        return "No Status";
                    else
                        return "Loading";

                case 0x00000100:        // Loaded or Validating
                    if ( p == 0 )
                        return "Loaded";
                    else
                        return "Validating";

                case 0x00000300:        // Validated
                    if ( status.PortIndex == null )
                        return "Validated";
                    else
                        return "Active";

                case 0x80010000:        // Scene Disposed
                    return "Disposed";

                case 0x80020000:        // Load Error
                    return "Load Error";

                default:
                    Console.WriteLine("SHOULD NOT COME HERE!!!");
                    break;
            }

            return "Undefined";
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Loaded'
        /// </summary>
        public static bool IsLoaded(uint status)
        {
            return (status & 0x00000100) == 0x00000100;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Validated'
        /// </summary>
        public static bool IsValidated(uint status)
        {
            return (status & 0x00000300) == 0x00000300;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Disposed'
        /// </summary>
        public static bool IsDisposed(uint status)
        {
            return (status & 0x80010000) == 0x80010000;
        }

        /// <summary>
        /// Returns true if 'status' value represents 'Error'
        /// </summary>
        public static bool IsError(uint status)
        {
            return (status & 0x80020000) == 0x80020000;
        }

        /// <summary>
        /// Returns a combined message text of the specified Exception
        /// and all of its inner Exceptions.
        /// </summary>
        public static string GetFullExceptionText(Exception ex)
        {
            if ( ex != null )
            {
                string msg = ex.Message;
                if ( ex.InnerException != null )
                    return msg + "\n" + GetFullExceptionText(ex.InnerException);
                else
                    return msg;
            }

            return string.Empty;
        }

        /// <summary>
        /// Returns a LogItem from a LogEventArgs
        /// </summary>
        public static LogItem GetLogItem(LogEventArgs le)
        {
            if ( le == null ) return null;

            return new LogItem(le.Level, le.Message, le.Module);
        }

        /// <summary>
        /// Creates SceneModel instance from the string represetation.
        /// </summary>
        /// <param name="sceneModelString">string representing the SceneModel to create</param>
        /// <param name="sceneIdentity">Identity of the scene the SceneModel belongs to</param>
        public static Ventuz.Remoting4.SceneModel.SceneModel GetSceneModel(string sceneModelString, string sceneIdentity)
        {
            if ( string.IsNullOrWhiteSpace(sceneModelString) )
                throw new ArgumentException("SceneModel string is empty!");

            if ( string.IsNullOrWhiteSpace(sceneIdentity) )
                throw new ArgumentException("Scene Identity string is empty!");

            Ventuz.Remoting4.SceneModel.SceneModel sm = null;
            using ( MemoryStream ms = new MemoryStream(Encoding.UTF8.GetBytes(sceneModelString ?? "")) )
            {
                sm = Ventuz.Remoting4.SceneModel.SceneModel.FromStream(ms, sceneIdentity);
            }

            return sm;
        }

        #endregion
    }

    #endregion
}
