﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Xml.Linq;
// Ventuz related namespaces
using Ventuz.Remoting4;
using Ventuz.Remoting4.MachineService;
using Ventuz.Remoting4.Templates;


namespace Remoting4_Demo_4
{
    // ------------------------------------------------------------------------------------------
    //
    // This Demo shows how to
    // 
    // 1. Use the VMSDiscovery class to get a list of Ventuz machines in the local network
    // 2. Start a Ventuz Project on a multi-machine Cluster
    // 3. List and load any available scenes in the configured Ventuz Project
    // 4. List available Templates in a scene
    // 5. Cue/Take/Clear Templates to individual Ports
    // 6. Retrieve TemplateData from a Template and fill it with values
    //
    // ------------------------------------------------------------------------------------------

    // TODO:
    // - handle not cue-able Templates (sub-Templates)

    // TEST:
    // Use Layout Scene with Template and try to cue this template!

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        #region Properties / Members

        /// <summary>
        /// Represents the remote connection to the Ventuz machine(s).
        /// This is the central instance for Ventuz Remoting 4!
        /// </summary>
        Cluster cluster = null;

        /// <summary>
        /// Ventuz Machine Service Discovery
        /// </summary>
        VMSDiscovery vmsDiscover = null;

        /// <summary>
        /// IID of scene on which properties are set via Data Items
        /// </summary>
        IID remoteScene = IID.Invalid;

        /// <summary>
        /// IID of that scene which is active in Pipe 0 and thus rendered
        /// </summary>
        IID activeScene = IID.Invalid;

        /// <summary>
        /// List of active VMS Clients for communication with Ventuz Machines
        /// </summary>
        List<VMSClient> activeVMSClients = new List<VMSClient>();

        /// <summary>
        /// Collection of Templates added by user from available scenes
        /// </summary>
        ProjectTemplates projectTmpls = new ProjectTemplates();

        List<VMS> allVMachines = new List<VMS>();
        /// <summary>
        /// Contains all Ventuz Machines found by VMSDiscovery
        /// </summary>
        public List<VMS> AllVMachines
        {
            get { return this.allVMachines; }
            set
            {
                this.allVMachines = value;
                OnPropertyChanged();
            }
        }

        ObservableCollection<VMS> clusterVMachines = new ObservableCollection<VMS>();
        /// <summary>
        /// Ventuz Machines selected to build a Cluster
        /// </summary>
        public ObservableCollection<VMS> ClusterVMachines { get { return this.clusterVMachines; } }

        bool clusterStarted = false;
        /// <summary>
        /// Indicates that the Cluster is started.
        /// </summary>
        public bool ClusterStarted
        {
            get { return this.clusterStarted; }
            set
            {
                this.clusterStarted = value;

                OnPropertyChanged();
                OnPropertyChanged("ClusterStopped");
            }
        }

        /// <summary>
        /// Indicates that the Cluster is stopped.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool ClusterStopped { get { return !this.ClusterStarted; } }

        /// <summary>
        /// Indicates that machines can be added to a potential Cluster.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanAddMachines { get { return !this.ClusterStarted; } }

        /// <summary>
        /// Indicates that a machines can be removed from a potential Cluster.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanRemoveMachines { get { return this.clusterVMachines.Count > 0 && !this.ClusterStarted; } }

        /// <summary>
        /// Indicates that a Cluster can be started.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanStartCluster { get { return this.clusterVMachines.Count > 0 && !this.ClusterStarted; } }

        /// <summary>
        /// Indicates that a Cluster can be stopped.
        /// (Mainly used for WPF binding.)
        /// </summary>
        public bool CanStopCluster { get { return this.clusterVMachines.Count > 0 && this.ClusterStarted; } }

        ObservableCollection<R4Helper.LogItem> logEntries = new ObservableCollection<R4Helper.LogItem>();
        /// <summary>
        /// Collection of log entires from Cluster machines and local Remote Client app.
        /// </summary>
        public ObservableCollection<R4Helper.LogItem> LogEntries { get { return this.logEntries; } }

        // TODO: use ObservableCollection?
        List<string> files = null;
        /// <summary>
        /// List of available files - here mainly: Scenes
        /// </summary>
        public List<string> Files { get { return this.files; } }

        ObservableCollection<R4Helper.IIDItem> loadedScenes = new ObservableCollection<R4Helper.IIDItem>();
        /// <summary>
        /// Collection of scenes loaded by the Cluster
        /// </summary>
        public ObservableCollection<R4Helper.IIDItem> LoadedScenes { get { return this.loadedScenes; } }

        ObservableCollection<Template> loadedTemplates = new ObservableCollection<Template>();
        /// <summary>
        /// Collection of Templates available to be cued/taken
        /// </summary>
        public ObservableCollection<Template> LoadedTemplates { get { return this.loadedTemplates; } }

        // TODO: use ObservableCollection?
        List<R4Helper.DataItem> diProperties = null;
        /// <summary>
        /// List of Data Item properties on selected scene or template
        /// </summary>
        public List<R4Helper.DataItem> DataItemProperties { get { return this.diProperties; } }

        // TODO: use ObservableCollection?
        List<R4Helper.DataItem> diEvents = null;
        /// <summary>
        /// List of Data Item events/methods on selected scene or template
        /// </summary>
        public List<R4Helper.DataItem> DataItemEvents { get { return this.diEvents; } }

        ObservableCollection<R4Helper.FreePortInfo> freePorts = new ObservableCollection<R4Helper.FreePortInfo>();
        /// <summary>
        /// Dynamic Collection of available Ports to load a scene to or cue&take a template
        /// </summary>
        public ObservableCollection<R4Helper.FreePortInfo> FreePorts { get { return this.freePorts; } }

        bool loadSceneCtxEnabled = true;
        /// <summary>
        /// Property to disable certain commands on loaded scenes
        /// </summary>
        public bool LoadedScenesCtxMenuEnabled
        {
            get { return this.loadSceneCtxEnabled; }
            set
            {
                this.loadSceneCtxEnabled = value;
                OnPropertyChanged();
            }
        }

        /// <summary>
        /// TakeInfo returned by the last Cue.
        /// </summary>
        Ventuz.Remoting4.TakeInfo activeTakeInfo = null;

        /// <summary>
        /// Information about the Port where the last Cue was performed.
        /// </summary>
        R4Helper.FreePortInfo activePortInfo = null;

        /// <summary>
        /// TemplateData for the last cued Template.
        /// </summary>
        TemplateData activeTD = null;

        /// <summary>
        /// TemplateData which is currently active in the Data Item Editor.
        /// </summary>
        TemplateData editorActiveTD = null;

        /// <summary>
        /// Number of possible Takes for last cued Template.
        /// </summary>
        int takeCounter = 0;

        bool cueEnabled;
        /// <summary>
        /// Specifies if cueing a Template can be performed.
        /// </summary>
        public bool CueEnabled
        {
            get { return this.cueEnabled; }
            set
            {
                this.cueEnabled = value;
                OnPropertyChanged();
            }
        }

        bool takeEnabled;
        /// <summary>
        /// Specifies if taking a Template can be performed.
        /// </summary>
        public bool TakeEnabled
        {
            get { return this.takeEnabled; }
            set
            {
                this.takeEnabled = value;
                OnPropertyChanged();
            }
        }

        string takeText = "Take";
        /// <summary>
        /// Text for the Take button
        /// </summary>
        public string TakeText
        {
            get { return this.takeText; }
            set
            {
                this.takeText = value;
                OnPropertyChanged();
            }
        }

        bool clearEnabled;
        /// <summary>
        /// Specifies if clearing a Port can be performed.
        /// </summary>
        public bool ClearEnabled
        {
            get { return this.clearEnabled; }
            set
            {
                this.clearEnabled = value;
                OnPropertyChanged();
            }
        }

        /// <summary>
        /// If True editor is showing Project Data items, if False Scene Data items and Null for Template Data.
        /// </summary>
        bool? isProjectData = false;

        /// <summary>
        /// Used to notify WPF bindings that the sources of the bindings have changed.
        /// </summary>
        private void NotifyPropertyChanges()
        {
            OnPropertyChanged("CanAddMachines");
            OnPropertyChanged("CanRemoveMachines");
            OnPropertyChanged("CanStartCluster");
            OnPropertyChanged("CanStopCluster");
        }

        #endregion


        #region Constructor / Close

        public MainWindow()
        {
            InitializeComponent();

            StartMachineDiscovery();
        }

        /// <summary>
        /// Window closed by user.
        /// </summary>
        private void Window_Closed(object sender, EventArgs e)
        {
            ShutDownCluster();
        }

        #endregion


        #region Ventuz Machine Discovery

        /// <summary>
        /// Starts Ventuz Machine Discovery to get infos about available Ventuz Machines in the LAN
        /// </summary>
        private void StartMachineDiscovery()
        {
            if ( this.vmsDiscover == null )
            {
                this.vmsDiscover = new VMSDiscovery(10, 21406, new int[] { 20406, 19406, 19402 }); // VMS Ports for V7, V6, V5, V4
                this.vmsDiscover.PropertyChanged += VMSDiscovery_Changed;
                this.vmsDiscover.Discover();
            }
            else
            {
                this.vmsDiscover.Discover();
            }
        }

        /// <summary>
        /// Handles the PropertyChanged event on VMSDiscovery class
        /// </summary>
        private void VMSDiscovery_Changed(object sender, PropertyChangedEventArgs e)
        {
            // create list from discovered Ventuz machines
            if ( this.allVMachines == null ) this.allVMachines = new List<VMS>();
            bool update = false;

            foreach ( var item in vmsDiscover.Items )
            {
                // check if list update is necessary
                if ( !this.allVMachines.Contains(item) )
                    update = true;
            }

            if ( update )
                this.AllVMachines = new List<VMS>(vmsDiscover.Items);
        }

        #endregion

        #region Cluster Control

        /// <summary>
        /// Handles adding of machine to potential Cluster
        /// </summary>
        private void AddMachineToCluster(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllMachines.SelectedItems.Count > 0 )
            {
                // add all selected items to list which defines the multi-machine Cluster.
                // double entries are handled in called method
                foreach ( var item in this.lvAllMachines.SelectedItems )
                {
                    AddToClusterList((VMS)item);
                }
            }
        }

        /// <summary>
        /// Double-click on machines adds it automatically to Cluster list
        /// </summary>
        private void MachineDoubleClick(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllMachines.SelectedItems.Count > 0 )
                AddToClusterList((VMS)this.lvAllMachines.SelectedItem);
        }

        #region AddToClusterList

        /// <summary>
        /// Adds IP of specified VMS instance to Cluster list if not already done.
        /// </summary>
        private void AddToClusterList(VMS vms)
        {
            if ( this.clusterVMachines.Count == 0 )
                this.clusterVMachines.Add(vms);
            else
            {
                // check if specified machine already in cluster list
                bool contained = false;
                foreach ( VMS item in this.clusterVMachines )
                {
                    if ( item.IPAddress.Equals(vms.IPAddress) )
                    {
                        contained = true;
                        break;
                    }
                }

                if ( !contained )
                    this.clusterVMachines.Add(vms);
            }

            // Notify WPF binding to update UI elements
            NotifyPropertyChanges();
        }

        #endregion

        /// <summary>
        /// Handles removing of machines from potential Cluster
        /// </summary>
        private void RemoveMachineFromCluster(object sender, RoutedEventArgs e)
        {
            if ( this.lvClusterMachines.SelectedItems.Count > 0 )
            {
                IList selected = this.lvClusterMachines.SelectedItems;
                for ( int i = selected.Count - 1; i >= 0; i-- )
                {
                    this.clusterVMachines.Remove((VMS)selected[i]);
                }

                // Notify WPF binding to update UI elements
                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Handles Cluster start command on the list of selected Ventuz machines
        /// </summary>
        private void StartCluster(object sender, RoutedEventArgs e)
        {
            // if Cluster not running yet, a Project must be specified
            if ( string.IsNullOrWhiteSpace(this.tbProject.Text) )
            {
                LogMessage("Cannot start Cluster. No Project specified!", LogLevel.Error, true);
                return;
            }

            if ( !this.ClusterStarted && this.clusterVMachines.Count > 0 )
            {
                // create Cluster and VMS Clients
                this.cluster = new Cluster { Name = "R4_Demo4" };
                this.activeVMSClients.Clear();

                foreach ( VMS vms in this.clusterVMachines )
                {
                    this.activeVMSClients.Add(new VMSClient(vms));
                    this.cluster.AddMachine(new IPEndPoint(vms.IPAddress, Cluster.DEFAULT_PORT));
                }

                // get info about VMS clients
                List<VentuzInfo> infos = new List<VentuzInfo>();
                List<int> notRunnig = new List<int>();

                for ( int i = 0; i < this.activeVMSClients.Count; i++ )
                {
                    try
                    {
                        infos.Add(this.activeVMSClients[i].Info());
                    }
                    catch ( VMSClientException vEx )
                    {
                        LogMessage("VMS Client not available", vEx, LogLevel.Error, false);  // just for info
                        notRunnig.Add(i);
                    }
                }

                // start Ventuz Runtimes with specified Project via VMS clients
                bool anyFails = false;
                bool restart = false;

                if ( notRunnig.Count != this.activeVMSClients.Count )
                {
                    // all machines are running a Ventuz application!
                    // stop and re-start with selected Project?
                    // aks what to do
                    MessageBoxResult result = MessageBox.Show("Some/All machines in the Cluster are already running a Ventuz Designer/Runtime! \nDo you want to restart them with the specified Project?", "Warning", MessageBoxButton.YesNo, MessageBoxImage.Warning);

                    if ( result == MessageBoxResult.Yes )
                    {
                        // kill Ventuz because it will be restarted afterwards
                        restart = true;

                        foreach ( VMSClient client in this.activeVMSClients )
                        {
                            try
                            {
                                client.Kill();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage($"Could not stop machine [{client.IPEndPoint.Address.ToString()}]", ex, LogLevel.Error, false);
                            }
                        }
                    }
                }
                else
                {
                    // none of the machines is running any Ventuz application
                    restart = true;
                }

                if ( restart )
                {
                    // try to start clients with specified Project
                    foreach ( VMSClient client in this.activeVMSClients )
                    {
                        try
                        {
                            client.Start(this.tbProject.Text);
                        }
                        catch ( Exception ex )
                        {
                            LogMessage($"Could not start machine [{client.IPEndPoint.Address.ToString()}]", ex, LogLevel.Error, false);
                            anyFails = true;
                        }
                    }
                }

                if ( anyFails )
                {
                    LogMessage("Some Machines of the Cluster could not be started. Check the log entries!", LogLevel.Error, true);
                }
                else
                {
                    // register Event Handler...
                    this.cluster.ClusterStateChanged += Cluster_ClusterStateChanged;
                    this.cluster.SceneStatusChanged += Cluster_SceneStatusChanged;
                    this.cluster.DataItemChanged += Cluster_DataItemChanged;
                    this.cluster.LogLevel = LogLevel.Debug;
                    this.cluster.Log += Cluster_Log;
                    // ... and finally start the Cluster
                    this.cluster.Start();
                    this.ClusterStarted = true;
                }

                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Handles Cluster stop command
        /// </summary>
        private void StopCluster(object sender, RoutedEventArgs e)
        {
            if ( this.activeVMSClients.Count > 0 )
            {
                // FIRST: shut down Cluster
                this.ShutDownCluster();

                // SECOND: stop Runtimes and disconnect from VMS clients
                foreach ( VMSClient client in this.activeVMSClients )
                {
                    try
                    {
                        client.Kill();
                    }
                    catch ( Exception ex )
                    {
                        LogMessage("Stopping Ventuz failed", ex, LogLevel.Error, false);
                    }

                    client.Disconnect();
                }

                // THIRD: clear lists
                this.activeVMSClients.Clear();

                if ( this.files != null )
                {
                    this.files.Clear();
                    this.files = null;
                }

                OnPropertyChanged("Files");

                this.loadedScenes.Clear();
                this.loadedTemplates.Clear();

                this.ClusterStarted = false;

                NotifyPropertyChanges();
            }
        }

        /// <summary>
        /// Shuts down the connection to a Ventuz instance.
        /// </summary>
        private void ShutDownCluster()
        {
            if ( this.cluster != null )
            {
                ClearDataItemsLists();

                // 'clear' ProjectTemplates
                this.projectTmpls = new ProjectTemplates();

                this.cluster.Shutdown();
                // un-register event handler
                this.cluster.ClusterStateChanged -= Cluster_ClusterStateChanged;
                this.cluster.SceneStatusChanged -= Cluster_SceneStatusChanged;
                this.cluster.DataItemChanged -= Cluster_DataItemChanged;
                this.cluster.Log -= Cluster_Log;
                // mark connection as dead
                this.cluster = null;
            }
        }

        /// <summary>
        /// Process Cluster state changes here
        /// </summary>
        private void HandleClusterStatusUpdate()
        {
            if ( this.cluster == null )
                return;

            if ( this.cluster.ClusterState == ClusterState.Ok )
            {
                // OK state:
                LogMessage("Cluster is ready!", LogLevel.Info, false);
                UpdateFiles("Scenes");
                UpdateLoadedScenes();
            }
            else if ( (this.cluster.ClusterState & ClusterState.ErrorMask) != 0 )
            {
                // Error state:
                LogMessage(string.Format("Cluster is in an Error state: {0}", this.cluster.ClusterState), LogLevel.Error, false);
            }
        }

        #endregion

        #region Cluster Event Handler

        /// <summary>
        /// Log entry is received from any of the Cluster machines
        /// </summary>
        void Cluster_Log(object sender, LogEventArgs e)
        {
            // Cluster log events must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(() => { this.logEntries.Add(R4Helper.GetLogItem(e)); });
        }

        /// <summary>
        /// Status of any of the loaded scenes has changed
        /// </summary>
        void Cluster_SceneStatusChanged(object sender, SceneStatusChangedEventArgs e)
        {
            // Scene state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(() => { HandleSceneStatusUpdate(e.IID, e.Status); });
        }

        /// <summary>
        /// Status of the Cluster has changed
        /// </summary>
        void Cluster_ClusterStateChanged(object sender, EventArgs e)
        {
            // Cluster state change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher
            Dispatcher.Invoke(() => { HandleClusterStatusUpdate(); });
        }

        /// <summary>
        /// Data Item changed: property value changed or event was fired
        /// </summary>
        void Cluster_DataItemChanged(object sender, DataItemChangedEventArgs e)
        {
            // Data Item change must be handled in UI thread (Main Thread):
            // using UI Thread Dispatcher            
            Dispatcher.Invoke(() => { HandleSceneDataChange(e.DataItem, e.IID, e.Value); });
        }

        #endregion

        #region Scene Control

        /// <summary>
        /// Handles load commands from the context menu on a scene
        /// </summary>
        private void SceneItemLoad_Click(object sender, RoutedEventArgs e)
        {
            MenuItem mi = sender as MenuItem;

            if ( mi != null )
            {
                if ( this.lvAllScenes.SelectedItems.Count > 0 )
                {
                    if ( mi.Name != "LoadTemplates" )
                    {
                        // get load modus
                        LoadFlags lf = LoadFlags.New | LoadFlags.LoadProgress;
                        switch ( mi.Name )
                        {
                            case "New":
                                break;
                            case "NewExisting":
                                lf |= LoadFlags.Existing;
                                break;
                            case "ExistingUnused":
                                lf = LoadFlags.Existing | LoadFlags.Unused;
                                break;
                        }

                        // load all selected scenes
                        foreach ( var item in this.lvAllScenes.SelectedItems )
                        {
                            LoadScene((string)item, lf);
                        }
                    }
                    else
                    {
                        // add Templates from selected scenes to ProjectTemplateHost
                        foreach ( var item in this.lvAllScenes.SelectedItems )
                        {
                            AddTemplatesFromScene((string)item);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Handles double-click on a scene: load existing or new instance of the selected scene
        /// </summary>
        private void SceneItem_DoubleClick(object sender, RoutedEventArgs e)
        {
            if ( this.lvAllScenes.SelectedItems.Count > 0 )
                LoadScene((string)this.lvAllScenes.SelectedItem, LoadFlags.New | LoadFlags.Existing);
        }

        /// <summary>
        /// Handles different commands on scenes available in the Cluster
        /// </summary>
        private async void SceneCommands_Click(object sender, RoutedEventArgs e)
        {
            MenuItem mi = sender as MenuItem;
            if ( mi != null )
            {
                if ( this.lvLoadedScenes.SelectedItems.Count > 0 )
                {
                    R4Helper.IIDItem idi = null;
                    R4Helper.FreePortInfo fpi = new R4Helper.FreePortInfo(0, "Default", 0);
                    string cmdName = mi.Name;

                    if ( mi.Header is R4Helper.FreePortInfo )
                    {
                        // get information about Port where the scene should be set from the Header of the MenuItem
                        cmdName = "Activate";
                        fpi = (R4Helper.FreePortInfo)mi.Header;
                    }

                    switch ( cmdName )
                    {
                        case "Validate":
                            // validate scene
                            try
                            {
                                List<Task> tasks = new List<Task>();

                                foreach ( var item in this.lvLoadedScenes.SelectedItems )
                                    tasks.Add(this.cluster.Validate(((R4Helper.IIDItem)item).IID, null));

                                await Task.WhenAll(tasks);
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Validation failed", ex, LogLevel.Error, true);
                            }

                            break;
                        case "Activate":
                            // activate scene in Port of Scene specified in the FreePortInfo class
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                await this.cluster.PortStatus(fpi.Host, fpi.PortIndex, true, idi.IID, null, null);
                                this.activeScene = idi.IID;
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Activation failed", ex, LogLevel.Error, true);
                            }

                            break;
                        case "Deactivate":
                            // remove scene from its assigned Port
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                if ( idi.FullStatus.PortIndex.HasValue )
                                    await this.cluster.PortStatus(idi.FullStatus.PortIID, idi.FullStatus.PortIndex.Value, null, null, null, null);
                                this.activeScene = IID.Invalid;
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Deactivation failed", ex, LogLevel.Error, true);
                            }

                            break;
                        case "Release":
                            // unload/dispose scene
                            try
                            {
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                // remove from port first
                                if ( idi.FullStatus.PortIndex.HasValue )
                                    await this.cluster.PortStatus(idi.FullStatus.PortIID, idi.FullStatus.PortIndex.Value, null, null, null, null);
                                this.activeScene = IID.Invalid;
                                // now dispose
                                await this.cluster.Release(idi.IID, null);

                                if ( idi.IID.Equals(this.remoteScene) )
                                    ClearDataItemsLists();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Release/Dispose failed", ex, LogLevel.Error, true);
                            }

                            break;
                        case "Refresh":
                            // refresh scene list 
                            this.UpdateLoadedScenes();

                            break;
                        case "Remote":
                            // make data items available in GUI for Remoting
                            try
                            {
                                StopDataItemReading();
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                this.tbSceneDITitle.Text = string.Format("Data Items for Scene: {0}", idi.Identity);
                                await ProcessSceneDataModelTree(idi, true);
                                StartDataItemReading();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Presenting Scene-Data items failed", ex, LogLevel.Error, true);
                            }

                            break;
                        case "RemoteTree":
                            // make scene tree data items available in GUI for Remoting
                            try
                            {
                                StopDataItemReading();
                                idi = (R4Helper.IIDItem)this.lvLoadedScenes.SelectedItem;
                                this.tbSceneDITitle.Text = string.Format("Data Items for Scene Tree beginning at: {0}", idi.Identity);
                                await ProcessSceneDataModelTree(idi, false);
                                StartDataItemReading();
                            }
                            catch ( Exception ex )
                            {
                                LogMessage("Presenting Scene-Data items failed", ex, LogLevel.Error, true);
                            }

                            break;
                    }
                }
            }
        }

        /// <summary>
        /// Handles the request of Project Data
        /// </summary>
        private async void ProjectData_Click(object sender, RoutedEventArgs e)
        {
            this.tbSceneDITitle.Text = "Data Items for Project Data";
            await ProcessProjectDataModel();
        }

        /// <summary>
        /// Handles different commands on a Template
        /// </summary>
        private void TemplateCommands_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                MenuItem mi = sender as MenuItem;

                if ( mi != null )
                {
                    // via Context Menu
                    if ( this.lvTemplates.SelectedItems.Count == 1 )
                    {
                        R4Helper.FreePortInfo fpi = null;
                        string cmdName = mi.Name;

                        if ( mi.Header is R4Helper.FreePortInfo )
                        {
                            // get information about Port where the Template schould be cued from the Header of the MenuItem
                            cmdName = "TCue";
                            fpi = (R4Helper.FreePortInfo)mi.Header;
                        }

                        PerformTemplateCommands(cmdName, fpi);
                    }
                }
                else
                {
                    // via Button
                    Button bt = sender as Button;

                    if ( bt != null )
                    {
                        R4Helper.FreePortInfo fpi = new R4Helper.FreePortInfo(0, "", 0); ;
                        string cmdName = "";
                        string btName = bt.Content.ToString().Substring(0, 3);

                        switch ( btName )
                        {
                            case "Cue": cmdName = "TCue"; break;
                            case "Tak": cmdName = "TTake"; break;
                            case "Cle": cmdName = "TClear"; break;
                            default:
                                break;
                        }

                        PerformTemplateCommands(cmdName, fpi);
                    }
                }
            }
            catch ( Exception ex )
            {
                LogMessage("Cue/Take/Clear failed", ex, LogLevel.Error, true);
            }
        }

        /// <summary>
        ///  Performs a Cue/Take/Clear on a Template or shows its Data Items.
        /// </summary>
        private async void PerformTemplateCommands(string command, R4Helper.FreePortInfo fpi)
        {
            if ( this.cluster == null )
                return;

            try
            {
                switch ( command )
                {
                    case "TCue":
                        Template t = this.lvTemplates.SelectedItem as Template;

                        // create TemplateData for selected Template; it's necessary for the Cue
                        if ( this.activeTD == null )
                            this.activeTD = new TemplateData(t);
                        else if ( this.activeTD.Template != t )
                            this.activeTD = new TemplateData(t);

                        this.activeTakeInfo = null;
                        this.activePortInfo = fpi;

                        UpdateTemplateDataFromEditor();

                        // cue on specific Port and Host and save TakeInfo
                        this.activeTakeInfo = await this.cluster.Cue(fpi.Host, fpi.PortIndex, TransitionOptions.Release, this.activeTD.ToString(Ventuz.Remoting4.Templates.DataFormat.BIN), Guid.NewGuid(), null, null); ;

                        // enable Take command
                        this.takeCounter = this.activeTakeInfo.Count;
                        this.TakeEnabled = true;
                        this.TakeText = $"Take ({this.activeTakeInfo[0].Duration} s)";      // Display legth of Take in seconds

                        UpdateLoadedScenes();

                        break;
                    case "TTake":
                        if ( this.takeCounter > 0 )
                        {
                            // perform Take of last cued Template
                            if ( this.activePortInfo != null )
                            {
                                await this.cluster.Take(this.activePortInfo.Host, this.activePortInfo.PortIndex, this.activeTakeInfo.TakeID, null, null);

                                EnableTemplateDataSet();
                            }

                            this.takeCounter--;

                            if ( this.takeCounter == 0 )
                            {
                                // disable further Take commands but leave active Template Data alive for eventual 'on-air' updates
                                this.TakeEnabled = false;
                                this.TakeText = "Take";
                            }
                            else
                            {
                                this.TakeText = $"Take ({this.activeTakeInfo[activeTakeInfo.Count - takeCounter].Duration} s)";      // Display legth of Take in seconds
                            }
                        }

                        break;
                    case "TClear":
                        if ( this.activePortInfo != null )
                        {
                            // Cueing with 'null' data will signal that the following Take should take out the scene and clear the according ScenePort.
                            this.activeTakeInfo = await this.cluster.Cue(this.activePortInfo.Host, this.activePortInfo.PortIndex, TransitionOptions.Release, null, Guid.NewGuid(), null, null);

                            this.TakeText = $"Take ({this.activeTakeInfo[0].Duration} s)";      // Display legth of Take in seconds

                            // This Take will finally clear the ScenePort
                            await this.cluster.Take(this.activePortInfo.Host, this.activePortInfo.PortIndex, this.activeTakeInfo.TakeID, null, null);

                            DisableTemplateDataSet();

                            this.activePortInfo = null;
                            this.takeCounter = 0;
                            this.TakeEnabled = false;
                            this.activeTD = null;
                        }

                        break;
                    case "TData":
                        ProcessDataModel((Template)this.lvTemplates.SelectedItem);
                        break;
                }
            }
            catch ( Exception ex )
            {
                LogMessage("Cue/Take/Clear/Get DataItems failed", ex, LogLevel.Error, true);
            }
        }

        /// <summary>
        /// Sets new value of the according Data Item on the according scene
        /// </summary>
        private void SceneItemValueSet_Click(object sender, RoutedEventArgs e)
        {
            // get Data Item wrapper - it's the DataContect of the clicked Button
            R4Helper.DataItem smi = ((Button)sender).DataContext as R4Helper.DataItem;

            if ( smi != null )
            {
                try
                {
                    SetDataValue(smi, null);
                }
                catch ( Exception ex )
                {
                    LogMessage("Setting SceneData failed", ex, LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles command to set all available input properties on a scene via Data Items
        /// </summary>
        private async void SceneItemValueSetAll_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null && this.diProperties != null )
            {
                // get Cluster Clock to schedule Data Item set commands to future
                long clock = 0;
                bool failed = true;

                while ( failed )
                {
                    try
                    {
                        clock = await this.cluster.Clock(null);
                        failed = false;
                    }
                    catch
                    {
                        LogMessage("#### Get Cluster Clock failed! #####", LogLevel.Debug, false);
                        // failed - try again
                        failed = true;
                    }
                }
                clock += 20;    // schedule 20 frame in the future

                try
                {
                    bool? setBatched = this.cbComBatched.IsChecked;   // change for different Set method

                    if ( setBatched == false )
                    {
                        List<Task> tasks = new List<Task>();

                        if ( this.isProjectData == false && !this.remoteScene.IsInvalid )
                        {
                            // Scene data items
                            foreach ( R4Helper.DataItem item in this.diProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    tasks.Add(this.cluster.DataItem(item.ParentScene.IID, item.AddressPath, item.Value, clock, null));
                            }
                        }
                        else if ( this.isProjectData == true )
                        {
                            // Project data items
                            foreach ( R4Helper.DataItem item in this.diProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    tasks.Add(this.cluster.ProjectDataItem(item.AddressPath, item.Value, clock, null));
                            }
                        }

                        try
                        {
                            await Task.WhenAll(tasks);
                        }
                        catch ( Exception ex )
                        {
                            LogMessage("Setting DataItems failed", ex, LogLevel.Error, true);
                        }
                    }
                    else
                    {
                        // batch commands to set data items in parallel!
                        long commitClock = clock;
                        clock = await this.cluster.Batch(null);     // clock is now the Batch ID!

                        if ( this.isProjectData == false && !this.remoteScene.IsInvalid )
                        {
                            // Scene data items
                            foreach ( R4Helper.DataItem item in this.diProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    SetDataValue(item, clock);
                            }
                        }
                        else if ( this.isProjectData == true )
                        {
                            // Project data items
                            foreach ( R4Helper.DataItem item in this.diProperties )
                            {
                                if ( item.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) )
                                    SetDataValue(item, clock);
                            }
                        }

                        // commit the data item set commands!
                        await this.cluster.Batch(clock, true, commitClock, null);
                    }
                }
                catch ( Exception ex )
                {
                    LogMessage("Setting Data Items failed", ex, LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles clear command on Data Item lists
        /// </summary>
        private void DataItemsClear_Click(object sender, RoutedEventArgs e)
        {
            ClearDataItemsLists();
        }

        /// <summary>
        /// Shows a Snapshot Viewer dialog to display snapshot created via Remoting
        /// </summary>
        private void Snapshot_Click(object sender, RoutedEventArgs e)
        {
            if ( this.cluster != null )
            {
                ScreenshotDisplay sd = new ScreenshotDisplay();
                // set Cluster and Pipe to receive snapshot from
                sd.SetCluster(this.cluster, 0);
                sd.ShowDialog();
            }
        }

        /// <summary>
        /// Stops polling read-only Data Items and clears the lists
        /// </summary>
        private async void ClearDataItemsLists()
        {
            StopDataItemReading();

            // wait for Data Item reading thread to stop but do not block Main Thread!
            await Task.Factory.StartNew(() =>
            {
                while ( this.readingDataItems )
                {
                    Task.Delay(10);
                }
            });

            if ( this.diProperties != null )
            {
                this.diProperties.Clear();
                this.diProperties = null;
                OnPropertyChanged("DataItemProperties");
            }

            if ( this.diEvents != null )
            {
                this.diEvents.Clear();
                this.diEvents = null;
                OnPropertyChanged("DataItemEvents");
            }
        }

        /// <summary>
        /// Sets value on Data Item specified by its adress.
        /// </summary>
        /// <param name="adress">Data Item adress</param>
        /// <param name="value">value to set via Data Item</param>
        /// <param name="when">Cluster clock time to set Data Item at; can be null</param>
        private async void SetDataValue(R4Helper.DataItem item, long? when)
        {
            // Set value on DataItem specified by address asynchronously.
            if ( this.cluster != null )
            {
                try
                {
                    string address = item.AddressPath;
                    string value = item.Value;

                    if ( this.isProjectData == false && !this.remoteScene.IsInvalid )
                    {
                        // Scene Data item
                        // The IID of the according scene must be specified.
                        IID scene = item.ParentScene.IID;
                        await this.cluster.DataItem(scene, address, value, when, null);
                    }
                    else if ( this.isProjectData == true )
                        // Project Data item
                        await this.cluster.ProjectDataItem(address, value, when, null);
                    else if ( this.isProjectData == null )
                    {
                        // Template Data item
                        if ( this.activeTD != null && this.editorActiveTD != null && this.activeTD.Template == this.editorActiveTD.Template && this.activePortInfo != null )
                        {
                            // NOTE: Fullpath property is URL encoded!
                            string dd = System.Net.WebUtility.UrlDecode(this.activeTD.Template.Fullpath);
                            dd = dd.Substring(dd.IndexOf('/') + 1);
                            string dataItemAdress = string.Format("{0}.{1}", dd, address);

                            await this.cluster.DataItem(this.activePortInfo.Host, this.activePortInfo.PortIndex, dataItemAdress, value, when, null);
                        }
                        else if ( this.activeTD != null )
                            this.activeTD[item.AddressPath].SetValue(value);
                    }
                }
                catch ( Exception ex )
                {
                    LogMessage("Setting Data Item failed", ex, LogLevel.Error, false);
                }
            }
        }

        /// <summary>
        /// Handles the change of a Data Item  - here only Events are processed.
        /// Output properties do not notify via Remoting about changes to reduce network traffic;
        /// their values must be polled (see StartDataItemReading() method).
        /// </summary>
        /// <param name="dataItemAdress">adress of changed Data Item</param>
        /// <param name="sceneIID">IID of scene this Data Item belongs to</param>
        /// <param name="value">new Data Item value</param>
        private void HandleSceneDataChange(string dataItemAdress, IID sceneIID, object value)
        {
            // TODO: what about the '/' ??? Events come with leading '/'; is this a BUG?
            if ( this.diEvents != null )
            {
                dataItemAdress = dataItemAdress.Trim('/');

                foreach ( R4Helper.DataItem item in this.diEvents )
                {
                    if ( item.AddressPath == dataItemAdress )
                        item.RaiseIncomingEvent();
                }
            }
        }

        /// <summary>
        /// Loads scene asynchronously and adds it to list of loaded scenes.
        /// </summary>
        /// <param name="scene">identity of scene to load</param>
        /// <param name="loadFlags">load modus</param>
        private async void LoadScene(string scene, LoadFlags loadFlags)
        {
            try
            {
                FlaggedIID fIID = await this.cluster.Load(0, scene, loadFlags, null);

                if ( fIID.IID.HasValue )
                {
                    // wrap and add to list of loaded scenes
                    R4Helper.IIDItem iidi = new R4Helper.IIDItem(fIID.IID.Value);

                    if ( loadFlags == LoadFlags.New )
                    {
                        // new scene instance - add to list!
                        this.loadedScenes.Add(iidi);

                        // Get scene status to receive SceneStatusChanged events
                        iidi.FullStatus = await this.cluster.Status(fIID.IID.Value, null);
                    }
                    else
                    {
                        // could be reuse of already loaded scene - check if already in list
                        bool exists = false;

                        foreach ( R4Helper.IIDItem item in this.loadedScenes )
                        {
                            if ( item.IID.Equals(iidi.IID) )
                            {
                                // found
                                exists = true;
                                break;
                            }
                        }

                        if ( !exists )
                        {
                            this.loadedScenes.Add(iidi);
                            iidi.FullStatus = await this.cluster.Status(fIID.IID.Value, null);
                        }
                    }
                }
            }
            catch ( Exception ex )
            {
                LogMessage($"Loading scene '{scene}' failed", ex, LogLevel.Error, true);
            }
        }

        /// <summary>
        /// Adds all Templates from scene with the specified name to list of available Templates.
        /// </summary>
        private async void AddTemplatesFromScene(string sceneName)
        {
            // get SceneModel string from Cluster for specified scene
            string smString = await this.cluster.SceneModel(sceneName, null);

            if ( string.IsNullOrWhiteSpace(smString) )
                throw new ArgumentException(string.Format("Could not retrieve SceneModel string for: {0}", sceneName));

            // create SceneModel instance from string representation
            Ventuz.Remoting4.SceneModel.SceneModel sm = R4Helper.GetSceneModel(smString, sceneName);

            if ( sm == null )
                throw new ArgumentException(string.Format("Could not create SceneModel instance from SceneModel string for: {0}", sceneName));

            // create Scene Template from SceneModel
            SceneTemplates sceneTmpl = new SceneTemplates(sm);

            // add to ProjectTemplates template host
            this.projectTmpls.AddScene(sceneTmpl);

            // clear list of available Templates - it will be rebuilt from actual state of ProjectTemplates instance
            this.loadedTemplates.Clear();

            foreach ( var item in this.projectTmpls.Hosts )
            {
                AddTemplatesRecursive(item);
            }
        }

        /// <summary>
        /// Moves recursivly through the TemplateHost tree and adds all found Templates to a list.
        /// </summary>
        /// <param name="tHost">the TemplateHost instance to be inspected for cuable Templates</param>
        private void AddTemplatesRecursive(TemplateHost tHost)
        {
            if ( tHost == null )
                return;

            if ( tHost.Hosts.Count > 0 )
            {
                foreach ( var thItem in tHost.Hosts )
                    AddTemplatesRecursive(thItem);
            }

            if ( tHost.Templates.Count > 0 )
            {
                foreach ( var tItem in tHost.Templates )
                {
                    // do not add System (and Scene) Templates
                    if ( !tItem.Flags.HasFlag(TemplateFlags.System) && !tItem.Flags.HasFlag(TemplateFlags.Scene) )
                    {
                        this.loadedTemplates.Add(tItem);
                    }
                    else if ( tItem.Flags.HasFlag(TemplateFlags.Scene) )
                    {
                        if ( tItem.Host.Items.Count == 1 )
                        {
                            // must be a Scene Template without Animation Template: we allow Cue & Take
                            this.loadedTemplates.Add(tItem);
                        }
                    }
                }
            }
        }

        #region Data Items Processing

        /// <summary>
        /// Writes values from Editor to according TemplateData
        /// </summary>
        private void UpdateTemplateDataFromEditor()
        {
            if ( this.activeTD != null && this.editorActiveTD != null && this.isProjectData == null )
            {
                // check if Template matches the one of the Editor
                if ( this.activeTD.Template == this.editorActiveTD.Template && this.diProperties != null )
                {
                    foreach ( R4Helper.DataItem item in this.diProperties )
                    {
                        if ( !item.IsFolder )
                        {
                            try
                            {
                                this.activeTD[item.AddressPath].SetValue(item.Value);
                            }
                            catch ( Exception ex )
                            {
                                LogMessage(string.Format("Failed to set DataItem '{0}' to '{1}'", item.AddressPath, item.Value != null ? item.Value : "<NULL>"), ex, LogLevel.Error, false);
                            }
                        }
                    }
                }
                else if ( this.activeTD != this.editorActiveTD )
                {
                    LogMessage("Template Date in Editor does not correspond to cued Template!", LogLevel.Warning, false);
                }
            }
        }

        /// <summary>
        /// Enables data-items to be individually assignable.
        /// </summary>
        private void EnableTemplateDataSet()
        {
            if ( this.activeTD != null && this.editorActiveTD != null && this.activeTD.Template == this.editorActiveTD.Template )
            {
                if ( this.diProperties != null && this.diProperties.Count > 0 )
                {
                    foreach ( var item in this.diProperties )
                        item.SetEnabled = true;
                }

                if ( this.diEvents != null && this.diEvents.Count > 0 )
                {
                    foreach ( var item in this.diEvents )
                        item.SetEnabled = true;
                }
            }
        }

        /// <summary>
        /// Disables data-items to be individually assignable.
        /// </summary>
        private void DisableTemplateDataSet()
        {
            if ( this.diProperties != null && this.diProperties.Count > 0 )
            {
                foreach ( var item in this.diProperties )
                    item.SetEnabled = false;
            }

            if ( this.diEvents != null && this.diEvents.Count > 0 )
            {
                foreach ( var item in this.diEvents )
                    item.SetEnabled = false;
            }
        }

        /// <summary>
        /// Reads the Scene Model for the scene (tree) identified by the IID parameter and creates
        /// lists of properties and events which can be written or read.
        /// This provides the possibility to communicate with a remote scene.
        /// </summary>
        /// <param name="iid">IIDItem of the scene to communicate with</param>
        /// <param name="onlySpecifiedScene">if false the SceneModel of the full scene tree starting at the specified scene is processed</param>
        /// <returns></returns>
        private async Task ProcessSceneDataModelTree(R4Helper.IIDItem iid, bool onlySpecifiedScene)
        {
            // get Scene Model XML
            if ( this.cluster != null )
            {
                // remember root scene
                this.remoteScene = iid.IID;
                // new list if we start scene tree traversing
                this.diProperties = new List<R4Helper.DataItem>();
                this.diEvents = new List<R4Helper.DataItem>();
                this.isProjectData = false;

                // get scene tree beginning at current scene
                R4Helper.SceneTreeItem sceneTree = null;

                if ( onlySpecifiedScene )
                    // single scene
                    sceneTree = new R4Helper.SceneTreeItem(iid);
                else
                    // full scene tree
                    sceneTree = await GetSceneTree(iid);

                await ProcessSceneDataModelTree(sceneTree);

                GroupAndSortDataItems(false);

                // read actual values - might not match default values
                ReadCurrentDataItems();

                // WPF binding notification
                OnPropertyChanged("DataItemProperties");
                OnPropertyChanged("DataItemEvents");

                // enable 'Set All' button
                this.btSetAll.IsEnabled = true;
            }
        }

        /// <summary>
        /// Iterates through the scenes of the specified scene tree and processes the according
        /// SceneModels to retrieve properties and events.
        /// </summary>
        /// <param name="sceneTreeItem">the scene tree build by ScenePorts</param>
        /// <returns></returns>
        private async Task ProcessSceneDataModelTree(R4Helper.SceneTreeItem sceneTreeItem)
        {
            // get Scene Model XML
            string sceneModel = await this.cluster.SceneModel(sceneTreeItem.Item.IID, null);
            XElement root = XElement.Parse(sceneModel);
            XNamespace xns = root.GetDefaultNamespace();
            XElement dm = root.Element(xns + "SceneModel.DataModel").Element(xns + "DataModel");

            // get properties and events/methods
            ProcessAdditionalDataModel(dm, sceneTreeItem.Item);

            // do it recursivly for children
            if ( sceneTreeItem.Children != null )
            {
                foreach ( var child in sceneTreeItem.Children )
                    await ProcessSceneDataModelTree(child);
            }
        }

        /// <summary>
        /// Returns the scene tree with the specified IIDItem representing the tree root.
        /// </summary>
        /// <param name="parentIId">the root of the scene tree</param>
        /// <returns></returns>
        private async Task<R4Helper.SceneTreeItem> GetSceneTree(R4Helper.IIDItem parentIId)
        {
            if ( this.cluster != null )
            {
                IID[] allScenes = await this.cluster.Scenes(parentIId.IID, true, null);
                List<R4Helper.IIDItem> allIIs = new List<R4Helper.IIDItem>();

                if ( allScenes != null && allScenes.Length > 0 )
                {
                    foreach ( var sceneIid in allScenes )
                    {
                        R4Helper.IIDItem subII = new R4Helper.IIDItem(sceneIid);
                        subII.FullStatus = await this.cluster.Status(subII.IID, null);
                        allIIs.Add(subII);
                    }

                    R4Helper.SceneTreeItem parent = new R4Helper.SceneTreeItem(parentIId);
                    ConstructSceneTree(allIIs, parent, 0);
                    return parent;
                }
            }

            return null;
        }

        /// <summary>
        /// Recursively constructs a scene tree from a list of all loaded scenes and a SceneTreeItem that represents the tree root.
        /// </summary>
        /// <param name="allIIds">list of all loaded scenes represented by an IIDItem</param>
        /// <param name="parent">current scene tree leave</param>
        /// <param name="treeLevel">current scene tree level</param>
        private void ConstructSceneTree(List<R4Helper.IIDItem> allIIds, R4Helper.SceneTreeItem parent, int treeLevel)
        {
            treeLevel++;
            List<R4Helper.IIDItem> iidChilds = new List<R4Helper.IIDItem>();

            // check if any scenes are loaded by current parent scene
            foreach ( var item in allIIds )
            {
                if ( item.FullStatus.PortIID.Equals(parent.Item.IID) )
                    iidChilds.Add(item);
            }

            // sort children by port index
            iidChilds.Sort(R4Helper.IIDItem.CompareByPortIndex);

            foreach ( var iChild in iidChilds )
            {
                // construct subtrees for child nodes
                iChild.SceneTreeLevel = treeLevel;
                R4Helper.SceneTreeItem child = new R4Helper.SceneTreeItem(iChild);
                parent.AddChild(child);
                ConstructSceneTree(allIIds, child, treeLevel);
            }
        }

        /// <summary>
        /// Reads the Project Model and creates lists of properties and events which can be written or read.
        /// </summary>
        private async Task ProcessProjectDataModel()
        {
            // get Project Model XML
            if ( this.cluster != null )
            {
                string projectModel = await this.cluster.DataModel(null);

                XElement root = XElement.Parse(projectModel);

                this.isProjectData = true;

                ProcessDataModel(root);
            }
        }

        /// <summary>
        /// Creates lists of properties and events from specified XML which can be written or read.
        /// </summary>
        /// <param name="dataModel">XML which specifies the DataModel</param>
        private void ProcessDataModel(XElement dm)
        {
            try
            {
                // get all Data Item entries for remoting
                var leaves = from e in dm.Descendants() where !e.Elements().Any() select e;

                this.diProperties = new List<R4Helper.DataItem>();
                this.diEvents = new List<R4Helper.DataItem>();

                foreach ( XElement x in leaves )
                {
                    // create Data Item wrapper
                    R4Helper.DataItem smi = new R4Helper.DataItem(x);

                    // do not add folders to lists - this will be done later 
                    if ( smi.Itemtype == R4Helper.DataItem.ItemTypes.Unknown
                        || smi.Itemtype == R4Helper.DataItem.ItemTypes.Group
                        || smi.Itemtype == R4Helper.DataItem.ItemTypes.Animation )
                        continue;

                    // separate properties and events
                    if ( !smi.IsTrigger )
                        this.diProperties.Add(smi);
                    else
                        this.diEvents.Add(smi);
                }

                // group and sort according to names of data items 
                GroupAndSortDataItems();

                // read actual values - might not match default values
                ReadCurrentDataItems();

                // WPF binding notification
                OnPropertyChanged("DataItemProperties");
                OnPropertyChanged("DataItemEvents");

                // enable 'Set All' button
                this.btSetAll.IsEnabled = true;
            }
            catch ( Exception ex )
            {
                LogMessage("Retrieving Scene Model failed", ex, LogLevel.Error, true);
            }
        }

        private void ProcessAdditionalDataModel(XElement dm, R4Helper.IIDItem parentScene)
        {
            if ( this.diProperties == null ) throw new NullReferenceException("Property list is null!");
            if ( this.diEvents == null ) throw new NullReferenceException("Event list is null!");

            try
            {
                List<R4Helper.DataItem> diP = new List<R4Helper.DataItem>();
                List<R4Helper.DataItem> diE = new List<R4Helper.DataItem>();

                // get all Data Item entries for remoting
                var leaves = from e in dm.Descendants() where !e.Elements().Any() select e;

                foreach ( XElement x in leaves )
                {
                    // create Data Item wrapper
                    R4Helper.DataItem smi = new R4Helper.DataItem(x);
                    smi.ParentScene = parentScene;
                    smi.SceneTreeLevel = parentScene.SceneTreeLevel;

                    // do not add folders to lists - this will be done later 
                    if ( smi.Itemtype == R4Helper.DataItem.ItemTypes.Unknown
                        || smi.Itemtype == R4Helper.DataItem.ItemTypes.Group
                        || smi.Itemtype == R4Helper.DataItem.ItemTypes.Animation )
                        continue;

                    // separate properties and events
                    if ( !smi.IsTrigger )
                        diP.Add(smi);
                    else
                        diE.Add(smi);
                }

                // sort and add to global lists
                if ( diP.Count > 0 )
                {
                    diP.Sort(R4Helper.DataItem.CompareByAddress);
                    this.diProperties.AddRange(diP);
                }
                if ( diE.Count > 0 )
                {
                    diE.Sort(R4Helper.DataItem.CompareByAddress);
                    this.diEvents.AddRange(diE);
                }
            }
            catch ( Exception ex )
            {
                LogMessage("Retrieving Scene Model failed", ex, LogLevel.Error, true);
            }
        }

        /// <summary>
        /// Reads the Data Model of the specified Template and creates
        /// lists of properties and events which can be written back to 
        /// TemplateData if this Template is cued.
        /// </summary>
        private void ProcessDataModel(Template templ)
        {
            this.isProjectData = null;

            // enable 'on-air update' on TemplateDate items if according Template is already taken?
            bool enableItemSet = true;
            if ( this.activeTD != null && this.activeTD.Template == templ && this.activePortInfo != null )
                enableItemSet = true;

            this.activeTD = new TemplateData(templ);
            this.editorActiveTD = this.activeTD;

            this.diProperties = new List<R4Helper.DataItem>();
            this.diEvents = new List<R4Helper.DataItem>();

            foreach ( var tdv in this.activeTD.GetDataStorage().GetValues<ITemplateDataValue>() )
            {
                R4Helper.DataItem dt = new R4Helper.DataItem(tdv.Value);

                // do not add folders to lists - this will be done later 
                if ( dt.Itemtype == R4Helper.DataItem.ItemTypes.Unknown || dt.Itemtype == R4Helper.DataItem.ItemTypes.Group )
                    continue;

                // separate properties and events
                if ( !dt.IsTrigger )
                    this.diProperties.Add(dt);
                else
                    this.diEvents.Add(dt);
            }

            // group and sort according to names of data items 
            GroupAndSortDataItems();

            // WPF binding notification
            OnPropertyChanged("DataItemProperties");
            OnPropertyChanged("DataItemEvents");

            // set editor title
            this.tbSceneDITitle.Text = string.Format("Data Items for Template: {0}", this.activeTD.Template.DisplayName);
            // disable 'Set All' button
            this.btSetAll.IsEnabled = false;

            if ( enableItemSet )
                EnableTemplateDataSet();
        }

        /// <summary>
        /// Sorts and groups the currently available data items.
        /// </summary>
        private void GroupAndSortDataItems(bool sortAllDataItems = true)
        {
            Dictionary<string, string> folders = null;
            Dictionary<string, string> scenes = null;

            if ( this.diProperties != null && this.diProperties.Count > 0 )
            {
                // sort and group
                if ( sortAllDataItems )
                    this.diProperties.Sort(R4Helper.DataItem.CompareByAddress);

                // create folder entries
                int pl = this.diProperties.Count;
                folders = new Dictionary<string, string>();
                scenes = new Dictionary<string, string>();

                // ..for properties
                for ( int i = 0; i < pl; i++ )
                {
                    string folder = this.diProperties[i].GetFolderAddress();
                    string scene = null;

                    R4Helper.IIDItem ps = this.diProperties[i].ParentScene;
                    if ( ps != null )
                    {
                        string portIID = "????";
                        if ( !ps.FullStatus.PortIID.Equals(IID.Invalid) ) portIID = ps.FullStatus.PortIID.ToString();
                        string portIndex = "?";
                        if ( ps.FullStatus.PortIndex != null ) portIndex = ps.FullStatus.PortIndex.Value.ToString();

                        scene = string.Format("{0} [{1}] @ Port [{2}]:{3}", ps.Identity, ps.IID.ToString(), portIID, portIndex);
                    }

                    if ( scene != null )
                    {
                        string s;
                        if ( !scenes.TryGetValue(scene, out s) )
                        {
                            // remember the new forlder entry
                            scenes.Add(scene, scene);
                            // now insert at correct index
                            this.diProperties.Insert(i, new R4Helper.DataItem(scene, true));
                            i++;
                            pl++;
                        }
                    }

                    if ( folder != null )
                    {
                        string f;
                        string ff = string.Format("{0}.{1}", scene, folder);
                        if ( !folders.TryGetValue(ff, out f) )
                        {
                            // remember the new forlder entry
                            folders.Add(ff, folder);
                            // now insert at correct index
                            this.diProperties.Insert(i, new R4Helper.DataItem(folder));
                            i++;
                            pl++;
                        }
                    }
                }
            }

            if ( this.diEvents != null && this.diEvents.Count > 0 )
            {
                // .. and for events
                if ( sortAllDataItems )
                    this.diEvents.Sort(R4Helper.DataItem.CompareByAddress);

                folders = new Dictionary<string, string>();             // could be null!
                scenes = new Dictionary<string, string>();

                int pl = this.diEvents.Count;

                for ( int i = 0; i < pl; i++ )
                {
                    string folder = this.diEvents[i].GetFolderAddress();
                    string scene = null;

                    R4Helper.IIDItem ps = this.diEvents[i].ParentScene;
                    if ( ps != null )
                    {
                        string portIID = "????";
                        if ( !ps.FullStatus.PortIID.Equals(IID.Invalid) ) portIID = ps.FullStatus.PortIID.ToString();
                        string portIndex = "?";
                        if ( ps.FullStatus.PortIndex != null ) portIndex = ps.FullStatus.PortIndex.Value.ToString();

                        scene = string.Format("{0} [{1}] @ Port [{2}]:{3}", ps.Identity, ps.IID.ToString(), portIID, portIndex);
                    }

                    if ( scene != null )
                    {
                        string s;
                        if ( !scenes.TryGetValue(scene, out s) )
                        {
                            // remember the new forlder entry
                            scenes.Add(scene, scene);
                            // now insert at correct index
                            this.diEvents.Insert(i, new R4Helper.DataItem(scene, true));
                            i++;
                            pl++;
                        }
                    }

                    if ( !string.IsNullOrWhiteSpace(folder) )
                    {
                        string f;
                        string ff = string.Format("{0}.{1}", scene, folder);
                        if ( !folders.TryGetValue(ff, out f) )
                        {
                            folders.Add(ff, folder);
                            this.diEvents.Insert(i, new R4Helper.DataItem(folder));
                            i++;
                            pl++;
                        }
                    }
                }
            }
        }

        #endregion

        /// <summary>
        /// Reads all Data Item values via Remoting and saves these values in the according Data Item wrappers.
        /// </summary>
        private async void ReadCurrentDataItems()
        {
            try
            {
                if ( this.diProperties != null )
                {
                    // create Task for every Data Item read
                    List<Task<object>> readTasks = new List<Task<object>>();

                    if ( this.isProjectData == false )
                    {
                        // Scene Data items
                        for ( int i = 0; i < this.diProperties.Count; i++ )
                        {
                            if ( this.diProperties[i].ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.R) )
                                // not all entries represent Data Items - we have also folders in the list!
                                readTasks.Add(this.cluster.DataItem(this.diProperties[i].ParentScene.IID, this.diProperties[i].AddressPath, null));
                        }
                    }
                    else if ( this.isProjectData == true )
                    {
                        // Project Data items
                        for ( int i = 0; i < this.diProperties.Count; i++ )
                        {
                            if ( this.diProperties[i].ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.R) )
                                // not all entries represent Data Items - we have also folders in the list!
                                readTasks.Add(this.cluster.ProjectDataItem(this.diProperties[i].AddressPath, null));
                        }
                    }

                    // wait until all Tasks are finished
                    await Task.WhenAll(readTasks);

                    // save returned values
                    int r = 0;  // result index
                    for ( int i = 0; i < this.diProperties.Count; i++ )
                    {
                        if ( this.diProperties[i].ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.R) )
                        {
                            if ( readTasks[r].Result != null )
                                this.diProperties[i].Value = readTasks[r].Result.ToString();
                            r++;
                        }
                    }
                }
            }
            catch ( Exception ex )
            {
                LogMessage("Get DataItems (parallel) failed", ex, LogLevel.Error, false);
            }
        }

        /// <summary>
        /// Indicates that Data Items should be read (output properties)
        /// </summary>
        bool readDataItems = false;
        /// <summary>
        /// Indicates that Data Items are currently read in a background thread
        /// </summary>
        bool readingDataItems = false;

        /// <summary>
        /// Starts a background thread to read read-only Data Items (output properties)
        /// </summary>
        private void StartDataItemReading()
        {
            this.readDataItems = false;

            if ( this.diProperties != null && this.diProperties.Count > 0 )
            {
                Task.Factory.StartNew(() =>
                {
                    this.readDataItems = true;
                    this.readingDataItems = true;

                    try
                    {
                        while ( this.readDataItems )
                        {
                            foreach ( R4Helper.DataItem item in this.diProperties )
                            {
                                if ( !this.readDataItems )  // reading still requested?
                                    break;

                                if ( item.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                                {
                                    try
                                    {
                                        // read Data Item values synchronously...
                                        object v = this.cluster.DataItem(item.ParentScene.IID, item.AddressPath, null).Result;
                                        // ...and save in its wrapper
                                        item.Value = v.ToString();
                                    }
                                    catch ( Exception ex )
                                    {
                                        LogMessage("Failed to read Data Item", ex, LogLevel.Error, false);
                                    }

                                    if ( !this.readDataItems )  // reading still requested
                                        break;
                                }
                            }

                            // wait 500 milliseconds before starting over
                            Task.Delay(500);
                        }
                    }
                    finally
                    {
                        // signal that thread is finished
                        this.readingDataItems = false;
                    }
                });
            }
        }

        /// <summary>
        /// Stops the background thread for reading Data Item values
        /// </summary>
        private void StopDataItemReading()
        {
            this.readDataItems = false;
        }

        /// <summary>
        /// Searches the list of loaded scenes and updates its status.
        /// </summary>
        /// <param name="iid"></param>
        /// <param name="status"></param>
        private void HandleSceneStatusUpdate(IID iid, SceneStatus status)
        {
            // find according IIDItem
            if ( this.loadedScenes != null & this.loadedScenes.Count > 0 )
            {
                foreach ( R4Helper.IIDItem idi in this.loadedScenes )
                {
                    if ( idi.IID.Equals(iid) )
                    {
                        // ...and update its status

                        // For progress output:
                        //uint prog;
                        //R4Helper.GetSceneStatusValues(status, out prog);
                        //Console.WriteLine("Progress: {0}", (float)prog / 2.55f);

                        idi.FullStatus = status;
                        break;
                    }
                }
            }
        }

        /// <summary>
        /// Requests list of available files on the Cluster.
        /// </summary>
        /// <param name="datapool">Datapool that specifies the type of files</param>
        private async void UpdateFiles(string datapool)
        {
            try
            {
                string[] entries = await this.cluster.Files(datapool, null);
                this.files = new List<string>();
                this.files.AddRange(entries);
                OnPropertyChanged("Files");
            }
            catch ( Exception )
            {
                files.Clear();
                OnPropertyChanged("Files");
            }
        }

        /// <summary>
        /// Refreshes the status of loaded scenes by querying the Cluster
        /// </summary>
        private async void UpdateLoadedScenes()
        {
            try
            {
                this.loadedScenes.Clear();

                bool withLayoutScenes = true;
                IID[] scenes = await this.cluster.Scenes(0, withLayoutScenes, null);

                foreach ( IID iid in scenes )
                {
                    // wrap IID for easier handling
                    R4Helper.IIDItem iidi = new R4Helper.IIDItem(iid);

                    // Get scene status to receive SceneStatusChanged events
                    iidi.FullStatus = await this.cluster.Status(iid, null);

                    this.loadedScenes.Add(iidi);
                }
            }
            catch ( Exception ex )
            {
                LogMessage("Update of scene states failed", ex, LogLevel.Error, false);
            }
        }

        /// <summary>
        /// Dynamically creates sub-menu for the 'Activate' command on a loaded scene.
        /// The entries will be the Ports where the scene can be set to.
        /// </summary>
        private async void LoadedScenes_ContextMenuOpening(object sender, ContextMenuEventArgs e)
        {
            ListBox selectedItems = sender as ListBox;

            if ( selectedItems != null && selectedItems.SelectedItems.Count > 0 )
            {
                try
                {
                    // clear the list of free ports because it will be filled in the following
                    this.freePorts.Clear();

                    // disable certain commands if a LayoutScene was selected
                    // TODO: does not work for active custom LayoutScenes
                    R4Helper.IIDItem selIID = selectedItems.SelectedItem as R4Helper.IIDItem;

                    if ( selIID != null && selIID.FullStatus.Status == 0 )
                    {
                        LoadedScenesCtxMenuEnabled = false;
                        return;
                    }

                    LoadedScenesCtxMenuEnabled = true;

                    // get not selected items because they may have Ports where the selected scene may be set by user
                    List<R4Helper.IIDItem> notSelected = new List<R4Helper.IIDItem>();
                    foreach ( R4Helper.IIDItem item in this.loadedScenes )
                    {
                        if ( selectedItems.SelectedItems.Contains(item) )
                            continue;

                        if ( item is R4Helper.IIDItem )
                            notSelected.Add(item);
                    }

                    if ( notSelected.Count > 0 )
                    {
                        // update PortInfo on the scenes
                        await Task.WhenAll(notSelected.Where((ii) => (ii.FullStatus.Status & 0x80000000) == 0).Select((item) =>
                        {
                            return cluster.Ports(item.IID, null).ContinueWith((t) =>
                            {
                                item.PortInfos = t.Result;
                                return new PortInfo[0];
                            });
                        }).ToArray());

                        // do this async. because this causes UI updates in Main Thread.
                        // do not request status of already disposed scenes!
                        await Task.WhenAll(notSelected.Where((ii) => (ii.FullStatus.Status & 0x80000000) == 0).Select((item) =>
                        {
                            return cluster.Status(item.IID, null).ContinueWith((t) =>
                            {
                                item.FullStatus = t.Result;
                            });
                        }).ToArray());

                        // get PipeInfo to retrieve IDs of Layoutscenes
                        int pipeCount = 0;
                        VentuzInfo? vzInfo = cluster.VentuzInfo;
                        if ( vzInfo.HasValue )
                            pipeCount = vzInfo.Value.PipeCount;

                        List<PipeInfo> pipeinfos = new List<PipeInfo>();

                        if ( pipeCount > 0 )
                        {
                            await Task.WhenAll(Enumerable.Range(0, pipeCount).Select(x =>
                            {
                                return cluster.PipeInfo(x, null).ContinueWith((t) =>
                                {
                                    pipeinfos.Add(t.Result);
                                });
                            }).ToArray());
                        }

                        #region Search available Ports

                        // list of free ports per pipe:
                        // every Scene Key in the Dict. referes to an array of Port Tuples which consist of Port Index and Name
                        Dictionary<IID, R4Helper.FreePortInfo[]> freePortsPerPipe = new Dictionary<IID, R4Helper.FreePortInfo[]>();
                        List<R4Helper.FreePortInfo[]> freePortsPipeless = new List<R4Helper.FreePortInfo[]>();
                        bool onlyDefaultLayouts = true;

                        // first we fill with all available (not locked) ports
                        foreach ( R4Helper.IIDItem iitem in notSelected )
                        {
                            if ( (iitem.FullStatus.Status & 0x80010000) != 0 )
                                // skip disposed scenes
                                continue;

                            if ( iitem.PortInfos != null && iitem.PortInfos.Length > 0 )
                            {
                                IID pipeIdx = iitem.IID;
                                bool isLayout = false;

                                // check if this is a Pipe-LayoutScene
                                for ( int i = 0; i < pipeinfos.Count; i++ )
                                {
                                    if ( pipeinfos[i].LayoutIID.Value.Equals(iitem.IID) )
                                    {
                                        onlyDefaultLayouts = false;
                                        pipeIdx = (IID)i;   // adapt pipe index
                                        isLayout = true;
                                        break;
                                    }
                                }

                                // search free ports
                                List<R4Helper.FreePortInfo> pis = new List<R4Helper.FreePortInfo>();
                                int pIdx = 0;

                                foreach ( PortInfo pi in iitem.PortInfos )
                                {
                                    if ( (pi.Flags & PortFlags.Locked) == 0 )
                                        pis.Add(new R4Helper.FreePortInfo(pIdx++, pi.Name, iitem.IID));
                                    else
                                        pIdx++; // Locked: skip index
                                }

                                // update FreePipeInfo array for Pipe-LayoutScene or normal scene with ScenePorts 
                                if ( pis.Count > 0 )
                                {
                                    if ( isLayout || iitem.FullStatus.PortIndex.HasValue )
                                    {
                                        if ( !freePortsPerPipe.ContainsKey(pipeIdx) )
                                            freePortsPerPipe[pipeIdx] = pis.ToArray();
                                    }
                                    else
                                        freePortsPipeless.Add(pis.ToArray());
                                }
                            }
                        }

                        #endregion

                        #region Create Menu Items for Free Ports

                        if ( selectedItems.SelectedItems.Count == 1 )    // only single selection can be set to Port!
                        {
                            // check if selection is valid for setting to Port
                            bool selectionValid = true;

                            if ( selectedItems.SelectedItem is R4Helper.IIDItem )
                            {
                                for ( int i = 0; i < pipeinfos.Count; i++ )
                                {
                                    if ( pipeinfos[i].LayoutIID.Value.Equals(((R4Helper.IIDItem)selectedItems.SelectedItems[0]).IID) )
                                    {
                                        // LayoutScenes not allowed to be set to Port!
                                        selectionValid = false;
                                        break;
                                    }
                                }
                            }

                            if ( selectionValid && (freePortsPerPipe.Count > 0 || freePortsPipeless.Count > 0 || onlyDefaultLayouts) )
                            {
                                if ( onlyDefaultLayouts )
                                {
                                    // create entries for default pipes
                                    for ( int pi = 0; pi < pipeinfos.Count; pi++ )
                                    {
                                        R4Helper.FreePortInfo fpi = new R4Helper.FreePortInfo(0, "Port 0", (IID)pi);
                                        this.freePorts.Add(fpi);
                                    }
                                }

                                if ( freePortsPipeless.Count > 0 )
                                {
                                    foreach ( var freePorts in freePortsPipeless )
                                    {
                                        foreach ( var item in freePorts )
                                            this.freePorts.Add(item);
                                    }
                                }

                                if ( freePortsPerPipe.Count > 0 )
                                {
                                    foreach ( KeyValuePair<IID, R4Helper.FreePortInfo[]> freePorts in freePortsPerPipe )
                                    {
                                        foreach ( var item in freePorts.Value )
                                            this.freePorts.Add(item);
                                    }
                                }

                                // let's see if there are already any scenes in the available Ports
                                await Task.WhenAll(this.freePorts.Select((item) =>
                                {
                                // NOTE: setIID parameter must be IID.Invalid here!
                                return cluster.PortStatus(item.Host, item.PortIndex, null, IID.Invalid, null, null).ContinueWith((t) =>
                                    {
                                        if ( t.Result.IID != null )
                                            item.SceneInPort = t.Result.IID.ToString();
                                        else
                                            item.SceneInPort = "<Free>";

                                        return item;
                                    });
                                }).ToArray());
                            }
                        }

                        #endregion

                    }
                }
                catch ( Exception ex )
                {
                    LogMessage("Retrieving Port status for Context Menu failed", ex, LogLevel.Error, true);
                }
            }
        }

        #endregion

        #region Logging

        /// <summary>
        /// Handles the Clear Log contect menu command.
        /// </summary>
        private void LogClear_Click(object sender, RoutedEventArgs e)
        {
            if ( this.logEntries != null )
                this.logEntries.Clear();
        }

        /// <summary>
        /// Writes a message to the Log Window and brings up an optional Message Box to inform the user.
        /// </summary>
        /// <param name="message">message test</param>
        /// <param name="ex">exception</param>
        /// <param name="level">log/info severity</param>
        /// <param name="showPopup">true if a message box should pop up</param>
        private void LogMessage(string message, Exception ex, LogLevel level, bool showPopup)
        {
            string msg = R4Helper.GetFullExceptionText(ex);
            LogMessage($"{message}: {msg}", level, showPopup);
        }

        /// <summary>
        /// Writes a message to the Log Window and brings up an optional Message Box to inform the user.
        /// </summary>
        /// <param name="message">message test</param>
        /// <param name="level">log/info severity</param>
        /// <param name="showPopup">true if a message box should pop up</param>
        private void LogMessage(string message, LogLevel level, bool showPopup)
        {
            if ( string.IsNullOrWhiteSpace(message) )
                return;

            if ( !this.Dispatcher.CheckAccess() )
            {
                // adding Log Entries can only be done on main thread
                Dispatcher.Invoke(new Action<string, LogLevel, bool>(LogMessage), message, level, showPopup);
                return;
            }

            R4Helper.LogItem li = new R4Helper.LogItem(level, message, "Remote Client");
            this.logEntries.Add(li);

            if ( showPopup )
            {
                string caption = "";
                MessageBoxImage icon = MessageBoxImage.Information;

                switch ( level )
                {
                    case LogLevel.Error:
                    case LogLevel.Fatal:
                        caption = "Error";
                        icon = MessageBoxImage.Error;
                        break;

                    case LogLevel.Info:
                    case LogLevel.Debug:
                    case LogLevel.None:
                        caption = "Info";
                        icon = MessageBoxImage.Information;
                        break;

                    case LogLevel.Warning:
                        caption = "Warning";
                        icon = MessageBoxImage.Warning;
                        break;

                    default:
                        break;
                }

                MessageBox.Show(this, message, caption, MessageBoxButton.OK, icon);
            }
        } 

        #endregion


        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                handler(this, new PropertyChangedEventArgs(name));
        }

        #endregion
    }

    #region class DataItemTemplateSelector

    /// <summary>
    /// This class decides which WPF Template should be used for Data Items
    /// </summary>
    public class DataItemTemplateSelector : DataTemplateSelector
    {
        public DataTemplate PropertyWriteTemplate { get; set; }
        public DataTemplate PropertyReadTemplate { get; set; }
        public DataTemplate TriggerWriteTemplate { get; set; }
        public DataTemplate TriggerReadTemplate { get; set; }
        public DataTemplate FolderTemplate { get; set; }
        public DataTemplate SceneFolderTemplate { get; set; }

        public override DataTemplate SelectTemplate(object item, DependencyObject container)
        {
            R4Helper.DataItem sdi = item as R4Helper.DataItem;

            if ( sdi != null )
            {
                if ( sdi.IsScene )
                    return SceneFolderTemplate;
                else if ( sdi.IsFolder )
                    return FolderTemplate;
                else if ( !sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.RW )
                    // write properties
                    return PropertyWriteTemplate;
                else if ( !sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                    // read properties
                    return PropertyReadTemplate;
                else if ( sdi.IsTrigger && sdi.ReadWrite.HasFlag(Ventuz.Remoting4.SceneModel.DataMode.W) ) // for RW Trigger we currently use the 'Write' Template
                    // trigger methods
                    return TriggerWriteTemplate;
                else if ( sdi.IsTrigger && sdi.ReadWrite == Ventuz.Remoting4.SceneModel.DataMode.R )
                    // visualise events
                    return TriggerReadTemplate;
            }

            return null;
        }
    }

    #endregion

    #region Remoting 4 Helper class

    /// <summary>
    /// Remoting 4 helper class for some common tasks.
    /// </summary>
    public class R4Helper
    {
        #region inner class LogItem
        
        /// <summary>
        /// This class represents a log message entity.
        /// </summary>
        public class LogItem
        {
            /// <summary>
            /// Severity of the log entry
            /// </summary>
            public LogLevel Level { get; private set; }
            /// <summary>
            /// Log message text
            /// </summary>
            public string   Message { get; private set; }
            /// <summary>
            /// Source of the Log entry
            /// </summary>
            public string   Module { get; private set; }

            public LogItem(LogLevel level, string message, string module)
            {
                this.Level = level;
                this.Message = message;
                this.Module = module;
            }
        } 

        #endregion

        #region inner class IIDItem

        /// <summary>
        /// This class is a wrapper for the Ventuz.Remoting4.IID struct to provide a convenient access to some status properties
        /// </summary>
        public class IIDItem: INotifyPropertyChanged
        {
            public IID IID { get; private set; }
            public PortInfo[] PortInfos { get; set; }

            SceneStatus fullStatus;
            public SceneStatus FullStatus
            {
                get { return fullStatus; }
                set { fullStatus = value; UpdateStatus(); }
            }

            string identity;
            public string Identity 
            { 
                get{ return identity; }
                private set { identity = value; OnPropertyChanged(); } 
            }

            string status;
            public string Status 
            { 
                get{ return status; }
                private set { status = value; OnPropertyChanged(); } 
            }

            string port;
            public string Port 
            { 
                get{ return port; }
                private set { port = value; OnPropertyChanged(); } 
            }

            public int SceneTreeLevel { get; set; }

            private void UpdateStatus()
            {
                if ( this.fullStatus != null )
                {
                    this.Identity = this.fullStatus.Identity;
                    this.Status = R4Helper.GetSceneStatusString(this.fullStatus);

                    if ( this.fullStatus.PortIndex.HasValue )
                        this.Port = string.Format("{0}:{1}", this.fullStatus.PortIID, this.fullStatus.PortIndex.Value);
                    else
                        this.Port = "";
                }
                else
                {
                    this.Identity = "";
                    this.Status = "";
                    this.Port = "";
                }
            }

            public IIDItem(IID iid)
            {
                this.IID = iid;
                this.Identity = ""; 
                this.Port = ""; 
                this.Status = "";
             
                UpdateStatus();
            }

            public override string ToString()
            {
                if ( fullStatus == null )
                    return IID.ToString();

                try
                {
                    return string.Format("{0} [{1}] is {2} @ Port {3}:{4}", fullStatus.Identity, IID.ToString(), Status, fullStatus.PortIID.ToString(), fullStatus.PortIndex);
                }
                catch ( Exception ex )
                {
                    Console.WriteLine(ex.Message);
                }

                return base.ToString();
            }

            /// <summary>
            /// Compares two IIDItems by their PortIndex. Used for sorting.
            /// </summary>
            public static int CompareByPortIndex(IIDItem iid1, IIDItem iid2)
            {
                if ( iid1 == null )
                {
                    if ( iid2 == null )
                        // If both are null, they're equal.  
                        return 0;
                    else
                        // If iid1 is null and iid2 is not null, iid2 is greater.  
                        return -1;
                }
                else
                {
                    // If iid1 is not null... 
                    if ( iid2 == null )
                        // ...and iid2 is null, iid1 is greater.
                        return 1;
                    else
                    {
                        if ( iid1.FullStatus == null && iid2.FullStatus == null )
                            // both are equal
                            return 0;
                        else if ( iid1.FullStatus != null && iid2.FullStatus == null )
                            // iid1 is greater
                            return 1;
                        else if ( iid1.FullStatus == null && iid2.FullStatus != null )
                            // iid2 is greater
                            return -1;
                        else if ( iid1.FullStatus.PortIndex != null && iid2.FullStatus.PortIndex == null )
                            // iid1 is greater
                            return 1;
                        else if ( iid1.FullStatus.PortIndex == null && iid2.FullStatus.PortIndex != null )
                            // iid2 is greater
                            return -1;
                        else // iid1.FullStatus != null && iid2.FullStatus != null && iid1.FullStatus.PortIndex != null && iid2.FullStatus.PortIndex != null
                            return iid1.FullStatus.PortIndex.Value - iid2.FullStatus.PortIndex.Value;
                    }
                }
            }

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }

        #endregion

        #region inner class DataItem

        /// <summary>
        /// This class is a wrapper for Ventuz Remoting4 Data Items to provide a convenient access to their properties.
        /// This is e.g. necessary for usage in WPF templates.
        /// </summary>
        [System.Diagnostics.DebuggerDisplay("{FullAddressPath} [{ParentScene.IID}]")]
        public class                              DataItem: INotifyPropertyChanged
        {
            #region enum ItemTypes
            
            /// <summary>
            /// Enum to identify according property type for a Data Item
            /// </summary>
            public enum ItemTypes
            {
                Unknown = 0,
                Group = 1,
                String = 2,
                Float = 3,
                Double = 4,
                Integer = 5,
                Boolean = 6,
                Color = 7,
                Enum = 8,
                Asset = 9,
                Trigger = 10,
                StringArray = 11,
                FloatArray = 12,
                DoubleArray = 13,
                IntegerArray = 14,
                BooleanArray = 15,
                ByteArray = 16,
                Animation = 30,
            } 

            #endregion


            #region Properties

            /// <summary>
            /// Name of the Data Item. This is the last part of the remote address.
            /// </summary>
            public string Name { get; private set; }
            /// <summary>
            /// Full address path of this Data Item. Needed to read or write the value of this Data Item.
            /// </summary>
            public string AddressPath { get; private set; }
            /// <summary>
            /// Address path string with optional leading scene identity.
            /// </summary>
            public string FullAddressPath
            {
                get
                {
                    if ( ParentScene == null ) return this.AddressPath;
                    return string.Format("{0}{1}", ParentScene.Identity, AddressPath);
                }
            }
            /// <summary>
            /// Optional Default value of this Data Item.
            /// </summary>
            public string Default { get; private set; }

            string val = null;
            /// <summary>
            /// Actual value of this Data Item.
            /// </summary>
            public string Value
            {
                get{ return this.val; }
                set
                {
                    this.val = value;
                    OnPropertyChanged();
                }
            }

            /// <summary>
            /// Read/Write access of this Data Item
            /// </summary>
            public Ventuz.Remoting4.SceneModel.DataMode ReadWrite { get; private set; }
            /// <summary>
            /// Type of the according property/event/method of this Data Item
            /// </summary>
            public ItemTypes Itemtype { get; private set; }
            /// <summary>
            /// Indicates that this Data Item represents a Trigger/Method in the Ventuz scene.
            /// </summary>
            public bool IsTrigger { get { return this.Itemtype == ItemTypes.Trigger; } }

            /// <summary>
            /// Indicates that this instance represents only a folder but not a real Data Item
            /// </summary>
            public bool IsFolder { get; private set; }
            /// <summary>
            /// Indicates that this instance represents only a Scene but not a real Data Item
            /// </summary>
            public bool IsScene { get; private set; }
            /// <summary>
            /// Optional IIDItem that represents the parent scene of this DataItem
            /// </summary>
            public IIDItem ParentScene { get; set; }
            /// <summary>
            /// The level within a scene tree at which this item is located
            /// </summary>
            public int SceneTreeLevel { get; set; }

            private int status;
            /// <summary>
            /// Property used to visualize events
            /// </summary>
            public int Status
            {
                get { return this.status;}
                set
                {
                    this.status = value;
                    OnPropertyChanged();
                }
            }

            private bool setEnabled;
            /// <summary>
            /// Specifies that this DataItem can be set individually (true for Scene Data Items; false for Template Data Items)
            /// </summary>
            public bool SetEnabled 
            {
                get { return setEnabled; }
                set
                { 
                    this.setEnabled = value;
                    OnPropertyChanged();
                }
            }

            #endregion

            #region Constructors

            public DataItem(string folderPath, bool isScene = false)
            {
                this.AddressPath = folderPath;
                this.IsFolder = true;
                this.IsScene = isScene;
            }

            public DataItem(XElement xElem)
            {
                if ( xElem == null )
                    throw new ArgumentNullException("XElement is null!");

                if ( xElem.Attribute("Name") != null )
                    this.Name = xElem.Attribute("Name").Value;
                else
                    throw new ArgumentNullException("Missing attribute Name!");

                if ( xElem.Attribute("Default") != null )   // optional
                {
                    this.Default = xElem.Attribute("Default").Value;
                    this.Value = this.Default;
                }

                this.Itemtype = GetItemType(xElem.Name.LocalName);

                this.AddressPath = GetPath(xElem);

                if ( this.Itemtype > ItemTypes.Group && this.Itemtype != ItemTypes.Animation)
                {
                    if ( xElem.Attribute("Mode") != null )
                        this.ReadWrite = GetRWMode(xElem.Attribute("Mode").Value);
                    else
                        throw new ArgumentNullException("Missing attribute Mode!"); 
                }

                this.SetEnabled = true;
            }

            public DataItem(ITemplateDataValue itdValue)
            {
                if ( itdValue == null )
                    throw new ArgumentNullException("ITemplateDataValue is null!");

                this.Name = itdValue.Name;
                this.Itemtype = GetItemType(itdValue.ValueType);
                this.AddressPath = itdValue.Fullpath;

                if ( this.Itemtype > ItemTypes.Group )
                {
                    this.ReadWrite = itdValue.DataItem.Mode;
                }

                // Data Items from Template cannot be set individually
                this.SetEnabled = false;
            } 

            #endregion

            /// <summary>
            /// Changes Status property temporarily to visualize incoming event
            /// </summary>
            public void RaiseIncomingEvent()
            {
                this.Status = 1;
                this.Status = 0;
            }

            /// <summary>
            /// Returns the address path of the folder this Data Item is subordinated
            /// </summary>
            public string GetFolderAddress()
            {
                int lastPIdx = this.AddressPath.LastIndexOf('.');
                string folder = string.Empty;

                if ( lastPIdx > 0 )
                {
                    if ( this.AddressPath.IndexOf('.') == 0 )
                        folder = this.AddressPath.Substring(1, lastPIdx - 1);
                    else
                        folder = this.AddressPath.Substring(0, lastPIdx);
                }
             
                return folder;
            }

            #region Static Helper Methods

            /// <summary>
            /// Compares two SceneDataItems by their addresses. Used for sorting.
            /// </summary>
            public static int CompareByAddress(DataItem sdi1, DataItem sdi2)
            {
                if ( sdi1 == null )
                {
                    if ( sdi2 == null )
                        // If both are null, they're equal.  
                        return 0;
                    else
                        // If sdi1 is null and sdi2 is not null, sdi2 is greater.  
                        return -1;
                }
                else
                {
                    // If sdi1 is not null... 
                    if ( sdi2 == null )
                        // ...and sdi2 is null, sdi1 is greater.
                        return 1;
                    else
                    {
                        // ...and sdi2 is not null; first compare the scene levels (in case of a scene tree) 
                        int r = sdi1.SceneTreeLevel - sdi2.SceneTreeLevel;

                        if ( r != 0 )
                            return r;

                        // same scene Level: compare scene IIDs
                        if ( sdi1.ParentScene != null && sdi2.ParentScene != null )
                            r = string.Compare(sdi1.ParentScene.IID.ToString(), sdi2.ParentScene.IID.ToString());

                        if ( r != 0 )
                            return r;

                        // same scene: compare scene data item address
                        r = string.Compare(sdi1.FullAddressPath, sdi2.FullAddressPath);

                        if ( r != 0 )
                            return r;

                        // in case of same adress compare names
                        return string.Compare(sdi1.Name, sdi2.Name);
                    }
                }
            }

            /// <summary>
            /// Returns RWModes enum value from string
            /// </summary>
            private static Ventuz.Remoting4.SceneModel.DataMode GetRWMode(string rwMode)
            {
                if ( string.IsNullOrWhiteSpace(rwMode) )
                    throw new ArgumentException("Empty string does not represent an valid Read/Write Mode!");

                Ventuz.Remoting4.SceneModel.DataMode rw;
                if ( Enum.TryParse<Ventuz.Remoting4.SceneModel.DataMode>(rwMode, out rw) )
                    return rw;

                throw new ArgumentException("Invalid Read/Write Mode!");
            }

            /// <summary>
            /// Returns ItemTypes enum value from string
            /// </summary>
            private static ItemTypes GetItemType(string itemtype)
            {
                if ( string.IsNullOrWhiteSpace(itemtype) )
                    throw new ArgumentException("Empty string does not represent an valid Item Type!");

                ItemTypes it;
                if ( Enum.TryParse<ItemTypes>(itemtype, out it) )
                    return it;

                throw new ArgumentException("Invalid Item Type!");
            }

            /// <summary>
            /// Returns ItemTypes enum value from Type 
            /// </summary>
            private static ItemTypes GetItemType(Type type)
            {
                if ( type == null )
                    throw new ArgumentException("Empty string does not represent an valid Item Type!");

                ItemTypes it = ItemTypes.Group;

                switch ( type.Name )
                {
                    case "DataStorage": it = ItemTypes.Group; break;
                    case "String": it = ItemTypes.String; break;
                    case "Int32": it = ItemTypes.Integer; break;
                    case "Single": it = ItemTypes.Float; break;
                    case "Double": it = ItemTypes.Double; break;
                    case "Boolean": it = ItemTypes.Boolean; break;
                    case "Color": it = ItemTypes.Color; break;
                    case "Uri": it = ItemTypes.Asset; break;
                    case "ICommand": it = ItemTypes.Trigger; break;
                    case "String[]": it = ItemTypes.StringArray; break;
                    case "Int32[]": it = ItemTypes.IntegerArray; break;
                    case "Single[]": it = ItemTypes.FloatArray; break;
                    case "Double[]": it = ItemTypes.FloatArray; break;
                    case "Boolean[]": it = ItemTypes.BooleanArray; break;
                    case "Byte[]": it = ItemTypes.ByteArray; break;

                    default:
                        break;
                }

                return it;
            }

            /// <summary>
            /// Returns address path for specified XElement
            /// </summary>
            private static string GetPath(XElement element)
            {
                XElement tmp = element;
                string path = string.Empty;

                while ( tmp != null && tmp.Attribute("Name") != null )
                {
                    path = "." + tmp.Attribute("Name").Value + path;
                    tmp = tmp.Parent;
                }

                return path;
            } 

            #endregion

            #region INotifyPropertyChanged Members

            public event PropertyChangedEventHandler PropertyChanged;

            protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName]string name = null)
            {
                PropertyChangedEventHandler handler = PropertyChanged;
                if ( handler != null && !string.IsNullOrWhiteSpace(name) )
                    handler(this, new PropertyChangedEventArgs(name));
            }

            #endregion
        }
        
        #endregion

        #region inner class FreePortInfo

        /// <summary>
        /// Class that bundles information about assignable Ports
        /// </summary>
        public class FreePortInfo
        {
            public IID Host;
            public string PortName;
            public int PortIndex;
            public string SceneInPort = "";

            internal FreePortInfo(int portIndex, string portName, IID host)
            {
                Host = host;
                PortName = portName;
                PortIndex = portIndex;
            }

            public override string ToString()
            {
                return string.Format("{0} [{1}] @ {2} ({3})", PortName, PortIndex, Host, SceneInPort);
            }
        }

        #endregion

        #region inner class SceneTreeItem
        
        /// <summary>
        /// Simple class to represent a tree of scenes connected by ScenePorts.
        /// </summary>
        public class SceneTreeItem
        {
            public IIDItem Item { get; private set; }
            public List<SceneTreeItem> Children { get; private set; }

            public SceneTreeItem(IIDItem item)
            {
                if ( item == null )
                    throw new NullReferenceException("Parent is Null!");
                this.Item = item;
            }

            public void AddChild(SceneTreeItem child)
            {
                if ( Children == null )
                    Children = new List<SceneTreeItem>();

                Children.Add(child);
            }

            public void AddChild(IIDItem child)
            {
                if ( Children == null )
                    Children = new List<SceneTreeItem>();

                Children.Add(new SceneTreeItem(child));
            }
        } 

        #endregion


        #region static Helper Methods

        /// <summary>
        /// Returns the SceneStatus.Status value and an optional loading/validation progress.
        /// </summary>
        public static uint GetSceneStatusValues(SceneStatus status, out uint progress)
        {
            progress = status.Status & 0xFF;
            return status.Status & 0xFFFFFF00;
        }

        /// <summary>
        /// Returns the string representation of a SceneStatus.
        /// </summary>
        public static string GetSceneStatusString(SceneStatus status)
        {
            uint p;
            switch ( GetSceneStatusValues(status, out p) )
            {
                case 0x0:               // No Status or Loading
                    if ( p == 0 )
                        return "No Status";
                    else
                        return "Loading";

                case 0x00000100:        // Loaded or Validating
                    if ( p == 0 )
                        return "Loaded";
                    else
                        return "Validating";

                case 0x00000300:        // Validated
                    if ( status.PortIndex == null )
                        return "Validated";
                    else
                        return "Active";

                case 0x80010000:        // Scene Disposed
                    return "Disposed";

                case 0x80020000:        // Load Error
                    return "Load Error";

                default:
                    Console.WriteLine("SHOULD NOT COME HERE!!!");
                    break;
            }

            return "Undefined";
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Loaded'
        /// </summary>
        public static bool IsLoaded(uint status)
        {
            return (status & 0x00000100) == 0x00000100;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Validated'
        /// </summary>
        public static bool IsValidated(uint status)
        {
            return (status & 0x00000300) == 0x00000300;
        }

        /// <summary>
        /// Returns true if 'status' value represents Scene 'Disposed'
        /// </summary>
        public static bool IsDisposed(uint status)
        {
            return (status & 0x80010000) == 0x80010000;
        }

        /// <summary>
        /// Returns true if 'status' value represents 'Error'
        /// </summary>
        public static bool IsError(uint status)
        {
            return (status & 0x80020000) == 0x80020000;
        }

        /// <summary>
        /// Returns a combined message text of the specified Exception
        /// and all of its inner Exceptions.
        /// </summary>
        public static string GetFullExceptionText(Exception ex)
        {
            if ( ex != null )
            {
                string msg = ex.Message;
                if ( ex.InnerException != null )
                    return msg + "\n" + GetFullExceptionText(ex.InnerException);
                else
                    return msg;
            }

            return string.Empty;
        }

        /// <summary>
        /// Returns a LogItem from a LogEventArgs
        /// </summary>
        public static LogItem GetLogItem(LogEventArgs le)
        {
            if ( le == null ) return null;

            return new LogItem(le.Level, le.Message, le.Module);
        }

        /// <summary>
        /// Creates SceneModel instance from the string represetation.
        /// </summary>
        /// <param name="sceneModelString">string representing the SceneModel to create</param>
        /// <param name="sceneIdentity">Identity of the scene the SceneModel belongs to</param>
        public static Ventuz.Remoting4.SceneModel.SceneModel GetSceneModel(string sceneModelString, string sceneIdentity)
        {
            if ( string.IsNullOrWhiteSpace(sceneModelString) )
                throw new ArgumentException("SceneModel string is empty!");

            if ( string.IsNullOrWhiteSpace(sceneIdentity) )
                throw new ArgumentException("Scene Identity string is empty!");

            Ventuz.Remoting4.SceneModel.SceneModel sm = null;
            using ( MemoryStream ms = new MemoryStream(Encoding.UTF8.GetBytes(sceneModelString ?? "")) )
            {
                sm = Ventuz.Remoting4.SceneModel.SceneModel.FromStream(ms, sceneIdentity);
            }

            return sm;
        }

        #endregion
    }

    #endregion
}
